import datajoint as dj


def schema(conn):

    schema = dj.Schema(
        schema_name="antelope_behaviour", connection=conn, create_tables=True
    )

    # import the metadata schema
    metadata = dj.create_virtual_module("metadata", "antelope_metadata")

    @schema
    class BehaviourRig(dj.Manual):
        definition = """
        # Custom json mapping for the behaviour rig
        -> metadata.Experimenter
        behaviourrig_id: smallint # Unique identifier for the behaviour rig (auto_increment)
        ---
        behaviourrig_name: varchar(40) # Name of the behaviour rig
        rig_json: json # Custom json for the behaviour rig
        behaviourrig_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class MaskFunction(dj.Manual):
        definition = """
        # Custom analysis function holding the masking functions for a rig
        -> BehaviourRig
        mask_id: smallint # Unique identifier for the mask function in the rig (auto_increment)
        ---
        maskfunction_name: varchar(40) # Name of the mask function
        mask_function: json # Serialised version of the mask function
        maskfunction_description: varchar(500) # Description of the mask function
        maskfunction_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class LabelledFrames(dj.Manual):
        definition = """
        # Holds the user labelled frames for deeplabcut training
        -> BehaviourRig
        -> metadata.Experiment
        dlcmodel_id: smallint # Unique identifier for the DLC model in the rig (auto_increment)
        ---
        dlcparams: json # Parameters for the DLC model
        labelled_frames: attach@labelled_frames # External data for the labelled frames
        labelledframes_in_compute: enum('False', 'True') # Are the labelled frames in a computation
        labelledframes_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class DLCModel(dj.Computed):
        definition = """
        # The trained DLC model for a particular rig
        -> BehaviourRig
        -> metadata.Experiment
        dlcmodel_id: smallint # Unique identifier for the DLC model in the rig (auto_increment)
        ---
        dlcmodel: attach@dlcmodel # External data for the DLC model
        evaluation_metrics: json # Evaluation metrics for the DLC model
        evaluated_frames: attach@evaluated_frames # Labelled images for validation
        dlcmodel_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Feature(dj.Manual):
        definition = """
        # Features for objects in the behaviour rig
        -> BehaviourRig
        feature_id: int # Unique identifier for a feature in the rig (auto_increment)
        ---
        feature_name: varchar(40) # Name of the feature
        source_type: enum('acquisition', 'stimulus', 'processing', 'deeplabcut') # Type of source for the feature
        data_type: enum('analog', 'digital', 'interval', 'kinematics') # Type of data for the feature
        feature_description: varchar(500) # Description of the feature
        feature_data=null: attach@feature_behaviour # External data for the feature
        feature_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class World(dj.Manual):
        definition = """
        # Represents the world for a particular session
        -> metadata.Session
        ---
        -> BehaviourRig
        dlc_training = 'False': enum('False', 'True') # Was the DLC model trained on this data
        world_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Video(dj.Imported):
        definition = """
        # Represents the video for a particular session
        -> World
        -> BehaviourRig
        video_id: smallint # Unique identifier for the video in the session (auto_increment)
        ---
        video: attach@behaviour_video # External data for the video
        timestamps = NULL: longblob # numpy array of event timestamps
        video_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Self(dj.Manual):
        definition = """
        # Represents the self for a particular session
        -> World
        -> metadata.Animal
        ---
        self_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Object(dj.Imported):
        definition = """
        # Represents the environment objects for a particular session
        -> World
        -> Feature
        ---
        object_name: varchar(40) # Name of the object
        object_type: enum('World', 'Self') # Type of object
        -> [nullable] Self
        object_deleted = 'False': enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class AnalogEvents(dj.Imported):
        definition = """
        # Represents the analog events for a particular session
        -> Object
        ---
        data: longblob # numpy array of event values
        -> [nullable] Self
        timestamps: longblob # numpy array of event timestamps
        x_coordinate: float # X coordinate of the feature in the rig
        y_coordinate: float # Y coordinate of the feature in the rig
        z_coordinate: float # Z coordinate of the feature in the rig
        unit: varchar(40) # units of the array
        analogevents_name: varchar(40) # Name of the analog event
        analogevents_deleted = 'False': enum('False', 'True')
        """

    @schema
    class DigitalEvents(dj.Imported):
        definition = """
        # Represents the digital events for a particular session
        -> Object
        ---
        data: longblob # numpy array of event values
        -> [nullable] Self
        timestamps: longblob # numpy array of event timestamps
        unit: varchar(40) # units of the array
        x_coordinate: float # X coordinate of the feature in the rig
        y_coordinate: float # Y coordinate of the feature in the rig
        z_coordinate: float # Z coordinate of the feature in the rig
        digitalevents_name: varchar(40) # Name of the digital event
        digitalevents_deleted = 'False': enum('False', 'True')
        """

    @schema
    class IntervalEvents(dj.Imported):
        definition = """
        # Represents the interval events for a particular session
        -> Object
        ---
        data: longblob # numpy array of event values
        -> [nullable] Self
        timestamps: longblob # numpy array of event timestamps
        x_coordinate: float # X coordinate of the feature in the rig
        y_coordinate: float # Y coordinate of the feature in the rig
        z_coordinate: float # Z coordinate of the feature in the rig
        intervalevents_name: varchar(40) # Name of the interval event
        intervalevents_deleted = 'False': enum('False', 'True')
        """

    @schema
    class Mask(dj.Imported):
        definition = """
        # Splits the recording session into individual trials
        -> World
        -> MaskFunction
        ---
        data: longblob # numpy array of event values
        timestamps: longblob # numpy array of event timestamps
        mask_name: varchar(40) # Name of the mask
        mask_deleted = 'False': enum('False', 'True')
        """

    @schema
    class Kinematics(dj.Computed):
        definition = """
        # Animal kinematics from DeepLabCut
        -> Object
        -> Video
        ---
        data: longblob # numpy array of kinematic values
        -> [nullable] DLCModel
        -> [nullable] Self
        timestamps: longblob # numpy array of kinematic timestamps
        kinematics_name: varchar(40) # Name of the kinematics
        kinematics_deleted = 'False': enum('False', 'True')
        """

    tables = {
        "BehaviourRig": BehaviourRig,
        "MaskFunction": MaskFunction,
        "LabelledFrames": LabelledFrames,
        "DLCModel": DLCModel,
        "Feature": Feature,
        "World": World,
        "Video": Video,
        "Self": Self,
        "Object": Object,
        "AnalogEvents": AnalogEvents,
        "DigitalEvents": DigitalEvents,
        "IntervalEvents": IntervalEvents,
        "Mask": Mask,
        "Kinematics": Kinematics,
    }

    return tables, schema



import streamlit as st
from antelope.utils.streamlit_utils import (
    dropdown_query_table,
    server_directory_browser,
)
from antelope.utils.antelope_utils import export_nwb
from antelope.utils.multithreading_utils import export_thread_pool
import pandas as pd


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Export Data")
        st.subheader("Select session")
        _, session = dropdown_query_table(
            tables, {"Session": tables["Session"]}, username, headless=True
        )

        if "session_id" in session.keys():
            st.subheader("Select download location")
            download_path = server_directory_browser("Select download location:")
            filename = st.text_input("Enter filename") + ".nwb"

            if st.button("Export"):

                export_pool = export_thread_pool()

                # submit job to thread pool
                future = export_pool.submit(
                    export_nwb, *(session, download_path / filename)
                )

                # append future to session state
                session_name = "_".join([str(i) for i in session.values()])
                st.session_state.export_futures.append((future, session_name))

                st.success("Export in progress.")

        if "export_futures" in st.session_state:
            if st.button("Check export progress"):

                st.write("Export statuses:")

                # initialise data
                display_futures = []

                # display job progress
                for future, query_name in st.session_state.export_futures:

                    # compute statuses
                    if future.done():
                        if future.exception():
                            print(future.exception())
                            status = "export error"
                        else:
                            status = "export success"
                    else:
                        status = "export in progress"

                    display_futures.append((query_name, status))

                # make dataframe to display
                df = pd.DataFrame(display_futures, columns=["Query", "Status"])

                # show dataframe
                st.dataframe(df, hide_index=True)



from antelope.connection import import_schemas
from antelope.utils.os_utils import get_config
from antelope.utils.analysis_base import AnalysisBase
from antelope.utils.hash_utils import dfs_functions, dfs_functions_not_tables
import importlib
import pkgutil
import inspect
import types
import sys
from pathlib import Path
import json
import ast


def import_analysis(conn, tables):

    # built in analysis functions
    analysis = importlib.import_module("antelope.analysis")
    functions = {}
    for _, name, _ in pkgutil.iter_modules(analysis.__path__):
        module = importlib.import_module(f"antelope.analysis.{name}")
        for _, cls in module.__dict__.items():
            if (
                inspect.isclass(cls)
                and issubclass(cls, AnalysisBase)
                and cls != AnalysisBase
            ):
                setattr(cls, "local", False)
                setattr(cls, "folder", name)
                if name not in functions:
                    functions[name] = []
                functions[name].append(cls(conn))

    # user defined analysis functions
    user_analysis = get_config()["analysis"]["folders"]
    for folder in user_analysis:
        sys.path.append(str(folder))
        for file in Path(folder).rglob("*.py"):
            module = importlib.import_module(file.stem)
            for _, cls in module.__dict__.items():
                if (
                    inspect.isclass(cls)
                    and issubclass(cls, AnalysisBase)
                    and cls != AnalysisBase
                ):
                    setattr(cls, "local", True)
                    setattr(cls, "folder", file.stem)
                    if file.stem not in functions:
                        functions[file.stem] = []
                    functions[file.stem].append(cls(conn))

    # add functions to functions
    new_functions = {}
    for folder_name, outer_folder in functions.items():
        new_functions[folder_name] = []
        for outer_function in outer_folder:

            # patch run method
            new_functions[folder_name].append(
                patch_functions(outer_function, functions, tables)
            )

    return new_functions


def instantiate_from_db(key, tables):

    # pull function definition from database
    function, name = (tables["AnalysisFunction"] & key).fetch1(
        "function", "analysisfunction_name"
    )

    # deserialise and instantiate
    function = json.loads(function)
    parsed = ast.parse(function)
    for node in ast.walk(parsed):
        if isinstance(node, ast.ClassDef):
            class_name = node.name
    exec(function)
    function = locals()[class_name](tables["Experimenter"].connection)
    function.local = False
    return function


def patch_functions(self, functions, tables):
    """
    This function patches the functions to allow other functions and tables to be accessible in the run() method.
    Has to be separate from the decorator as functions haven't yet been initialised with connections when the decorator is called.
    """

    new_globals = self.run.__globals__.copy()

    # add tables that are defined in function
    new_tables = []
    if isinstance(self.query, str):
        new_tables.append(self.query)
    elif isinstance(self.query, list):
        new_tables = self.query
    if hasattr(self, "data"):
        if isinstance(self.data, str):
            new_tables.append(self.data)
        elif isinstance(self.data, list):
            new_tables += self.data
    for key in new_tables:
        new_globals[key] = tables[key]

    # add functions defined in calls
    if hasattr(self, "calls"):
        for fct in self.calls:
            if "." in fct:
                folder, function = fct.split(".")
                if folder not in new_globals:
                    new_globals[folder] = types.SimpleNamespace()
                for fct_inner in functions[folder]:
                    if fct_inner.name == function:
                        setattr(new_globals[folder], function, fct_inner)
            else:
                folder = Path(inspect.getsourcefile(self.__class__.__bases__[0])).stem
                for fct_inner in functions[folder]:
                    if fct_inner.name == fct:
                        new_globals[fct] = fct_inner

    new_run = types.FunctionType(
        self.run.__code__,
        new_globals,
        self.run.__name__,
        self.run.__defaults__,
        self.run.__closure__,
    )
    self.run = new_run

    # compute all tables to hash for later
    hash = dfs_functions(self, functions)
    self.hash = hash

    return self


def split_trials(data, mask):
    """
    Split trials based on mask

    Inputs:
    mask: tuple of data, timestamps
    data: tuple of data, timestamps

    Returns:
    list of tuples of data, timestamps
    """

    mask_data, mask_time = mask
    data_data, data_time = data

    start_times = mask_time[mask_data == 1]
    stop_times = mask_time[mask_data == -1]

    mask = (data_time[:, None] >= start_times) & (data_time[:, None] <= stop_times)
    trials = [
        (data_data[mask[:, i]], data_time[mask[:, i]]) for i in range(mask.shape[1])
    ]

    return trials


def instantiate_function(key, tables):
    """
    Function loads an analysis function from the masks table and returns an instantiation of it
    """
    # pull function definition from database
    function, name = (tables["MaskFunction"] & key).fetch1(
        "mask_function", "maskfunction_name"
    )

    # deserialise and instantiate
    function = json.loads(function)
    exec(function)
    function = locals()[name](tables["Experimenter"].connection)

    # patch functions to function
    functions = import_analysis(tables["Experimenter"].connection, tables)
    for folder_name, outer_folder in functions.items():
        function = patch_functions(function, functions, tables)

    return function


def get_docstring(function_string, name):
    """
    Gets docstring from function string
    """

    exec(json.loads(function_string))
    function = locals()[name]
    docstring = function.__doc__
    return docstring


def reload_analysis(conn, tables):
    """
    Reloads analysis functions and repatches them if the underlying script has changed
    """

    analysis = importlib.import_module("antelope.analysis")
    functions = {}
    for _, name, _ in pkgutil.iter_modules(analysis.__path__):
        module_name = f"antelope.analysis.{name}"
        if module_name in sys.modules:
            # reload module if it's already loaded
            module = importlib.reload(sys.modules[module_name])
        else:
            module = importlib.import_module(module_name)

        for _, cls in module.__dict__.items():
            if (
                inspect.isclass(cls)
                and issubclass(cls, AnalysisBase)
                and cls != AnalysisBase
            ):
                if name not in functions:
                    functions[name] = []
                functions[name].append(cls(conn))

    # user-defined analysis functions
    user_analysis = get_config()["analysis"]["folders"]
    for folder in user_analysis:
        sys.path.append(str(folder))
        for file in Path(folder).iterdir():
            if file.suffix != ".py":
                continue
            module_name = file.stem
            if module_name in sys.modules:
                # reload module if it's already loaded
                module = importlib.reload(sys.modules[module_name])
            else:
                module = importlib.import_module(module_name)

            for _, cls in module.__dict__.items():
                if (
                    inspect.isclass(cls)
                    and issubclass(cls, AnalysisBase)
                    and cls != AnalysisBase
                ):
                    if module_name not in functions:
                        functions[module_name] = []
                    functions[module_name].append(cls(conn))

    # add functions to new_functions with patched run method
    new_functions = {}
    for folder_name, outer_folder in functions.items():
        new_functions[folder_name] = []
        for outer_function in outer_folder:
            # patch run method
            new_functions[folder_name].append(
                patch_functions(outer_function, functions, tables)
            )

    return new_functions


def publish_function(function, analysis_functions, username, tables):

    admin_tables = import_schemas.schema_admin(tables["Experimenter"].connection)

    # first check all functions called by this function
    functions = dfs_functions_not_tables(function, analysis_functions)
    # hacky change this
    if len(functions) == 0:
        functions = [function]

    # publish all functions
    new = 0
    with tables["Experimenter"].connection.transaction:
        for fct in functions:

            # make key
            insert_dict = {}
            insert_dict["experimenter"] = username
            insert_dict["analysisfunction_name"] = fct.name
            insert_dict["folder"] = fct.folder

            # check if in database
            if tables["AnalysisFunction"] & insert_dict:
                update = True
                insert_dict["analysisfunction_id"] = (
                    tables["AnalysisFunction"] & insert_dict
                ).fetch1("analysisfunction_id")
            else:
                update = False
                new += 1
                insert_dict["analysisfunction_id"] = (
                    max(
                        (tables["AnalysisFunction"] & {"experimenter": username}).fetch(
                            "analysisfunction_id"
                        ),
                        default=1,
                    )
                    + new
                )

            # add other attributes
            insert_dict["function"] = json.dumps(
                inspect.getsource(fct.__class__.__bases__[0])
            )
            if hasattr(fct, "hidden"):
                if fct.hidden:
                    insert_dict["hidden"] = "True"
                else:
                    insert_dict["hidden"] = "False"
            else:
                insert_dict["hidden"] = "False"
            insert_dict["analysisfunction_description"] = fct.__doc__

            if update:
                tables["AnalysisFunction"].update1(insert_dict)
            else:
                tables["AnalysisFunction"].insert1(insert_dict)



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.multithreading_utils import dlc_thread_pool, dlcup_thread_pool
from antelope.utils.antelope_utils import (
    create_dlc_folder,
    edit_params,
    check_dlc_folder,
    extract_frames,
    upload_dlc,
)
import streamlit_antd_components as sac
import pandas as pd
import json
import yaml


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Label frames for DeepLabCut training.")

        if len(tables["World"] & {"experimenter": username}) == 0:
            st.text("")
            st.warning("No data found.")
            return

        st.divider()

        st.subheader("Select Experiment")

        _, key = dropdown_query_table(
            tables, {"Experiment": tables["Experiment"]}, username, headless=True
        )
        if key:
            key["behaviourrig_deleted"] = "False"
        rigs = (tables["BehaviourRig"] & key).fetch()
        rigdict = {i["behaviourrig_name"]: i["behaviourrig_id"] for i in rigs}
        key["behaviourrig_id"] = rigdict[
            st.selectbox("Select behaviourrig_name", rigdict.keys())
        ]
        dlcmodel_id = (
            max(list((tables["LabelledFrames"] & key).fetch("dlcmodel_id")), default=0)
            + 1
        )
        st.selectbox("Select dlcmodel_id", [dlcmodel_id])
        key["dlcmodel_id"] = dlcmodel_id
        del key["experiment_deleted"]
        st.info(
            "You can only label frames for a new model here. If you want to update a previous training set, you need to delete it and create a new training dataset."
        )

        # fetch rig to check it has body parts
        rigjson = (tables["BehaviourRig"] & key).fetch1("rig_json")
        bodypart = False
        for feature in rigjson["features"]:
            if (
                "source_type" in feature["source"]
                and feature["source"]["source_type"] == "deeplabcut"
            ):
                bodypart = True

        if not bodypart:
            st.error(
                "This rig does not have any body parts to label. Please add body parts to the rig before proceeding."
            )
        else:

            # add dlc folder to session state if it exists
            if "dlc_folder" not in st.session_state:
                st.session_state.dlc_folder = {}
            status, conf = check_dlc_folder(key, tables)
            name = f"{key['experimenter']}_{key['behaviourrig_id']}_{key['experiment_id']}_{key['dlcmodel_id']}"
            if status:
                st.session_state.dlc_folder[name] = conf

            st.divider()

            sac.buttons(
                [
                    sac.ButtonsItem(label="Select Training Videos"),
                    sac.ButtonsItem(label="Download Videos"),
                    sac.ButtonsItem(label="Enter Parameters"),
                    sac.ButtonsItem(label="Extract Frames"),
                    sac.ButtonsItem(label="Label Frames"),
                ],
                key="mode",
                gap="md",
                use_container_width=True,
            )

            if st.session_state.mode == "Select Training Videos":

                videos = pd.DataFrame(
                    (tables["World"] * tables["Session"].proj("session_name") & key)
                    .proj("session_id", "session_name", "dlc_training")
                    .fetch()
                )
                videos["dlc_training"] = videos["dlc_training"].map(
                    {"True": True, "False": False}
                )

                st.text("")
                st.info(
                    "Here you can update which sessions belong in the training set for your deeplabcut model."
                )
                st.warning(
                    """Note, these videos will need to get downloaded to your local machine for you to annotate them, so selecting too many will make this download very slow."""
                )
                st.text("")

                edited_videos = st.data_editor(
                    videos,
                    column_config={"dlc_training": st.column_config.CheckboxColumn()},
                    disabled=["session_id", "session_name"],
                    column_order=["session_id", "session_name", "dlc_training"],
                    hide_index=True,
                    use_container_width=True,
                )

                st.text("")

                if st.button("Save"):

                    merge = pd.merge(edited_videos, videos, how="left", indicator=True)
                    merge = merge[merge["_merge"] == "left_only"]
                    merge["dlc_training"] = merge["dlc_training"].map(
                        {True: "True", False: "False"}
                    )
                    del merge["session_name"]
                    del merge["_merge"]

                    with tables["World"].connection.transaction:
                        for i, row in merge.iterrows():
                            tables["World"].update1(row.to_dict())

            if st.session_state.mode == "Download Videos":

                if name in st.session_state.dlc_folder.keys():
                    st.text("")
                    st.info("Training videos already downloaded.")

                else:

                    st.text("")
                    st.info(
                        "Training videos can take a while to download, so this will occur in a separate thread."
                    )
                    st.text("")

                    if st.button("Download training videos"):

                        dlc_pool = dlc_thread_pool()
                        future = dlc_pool.submit(create_dlc_folder, key)
                        st.session_state.download_dlc_futures.append((future, name))
                        st.success("Download in progress.")

                    # add a button which shows download statuses
                    # uses all downloads in current session stored in session state
                    if "download_dlc_futures" in st.session_state:

                        if st.button("Check download progress"):

                            st.write("Download statuses:")

                            # initialise data
                            display_futures = []

                            # display job progress
                            for (
                                future,
                                query_name,
                            ) in st.session_state.download_dlc_futures:

                                # compute statuses
                                if future.done():
                                    if future.exception():
                                        status = "download error"
                                        print(future.exception())
                                    else:
                                        status = "download success"
                                else:
                                    status = "download in progress"

                                display_futures.append((query_name, status))

                            # make dataframe to display
                            df = pd.DataFrame(
                                display_futures, columns=["Query", "Status"]
                            )

                            # show dataframe
                            st.dataframe(df, hide_index=True)

            if st.session_state.mode == "Enter Parameters":

                if name not in st.session_state.dlc_folder.keys():

                    st.error(
                        """You need to download the training videos to your machine first."""
                    )

                else:

                    st.text("")
                    st.info("Please edit the parameters for your DeepLabCut model.")
                    st.text("")
                    new_params = edit_params(key, tables)

                    st.text("")
                    if st.button("Save parameters"):
                        conf = st.session_state.dlc_folder[name]
                        with open(conf, "w") as f:
                            yaml.dump(new_params["config"], f)
                        comp = conf.parent / "compute.json"
                        with open(comp, "w") as f:
                            json.dump(new_params["compute"], f)
                        st.success("Parameters saved!")

            if st.session_state.mode == "Extract Frames":

                if name not in st.session_state.dlc_folder.keys():
                    st.error("Please enter parameters first.")
                    return
                else:
                    algo = st.selectbox(
                        "Select mode for frame extraction", ["uniform", "kmeans"]
                    )
                    st.text("")
                    if st.button("Extract frames"):
                        dlc_pool = dlc_thread_pool()
                        conf = st.session_state.dlc_folder[name]
                        future = dlc_pool.submit(extract_frames, *(conf, algo))
                        st.session_state.extract_frames_futures.append((future, name))
                        st.success("Extraction in progress.")

            if st.session_state.mode == "Label Frames":

                if name not in st.session_state.dlc_folder.keys():
                    st.error("Please enter parameters first.")

                elif not any(
                    (
                        st.session_state.dlc_folder[name].parent / "labeled-data"
                    ).iterdir()
                ):
                    st.error("Please extract frames first.")

                else:

                    conf = st.session_state.dlc_folder[name]
                    videos = conf.parent / "labeled-data"
                    vid_dict = {i.stem: i for i in videos.iterdir() if i.is_dir()}
                    selected_vid = vid_dict[
                        st.selectbox("Select video to label", list(vid_dict.keys()))
                    ]

                    st.text("")
                    st.info(
                        """Please use the button below to open a separate GUI to label your training frames. Make sure to save your results when you're done."""
                    )
                    st.text("")
                    if st.button("Label frames"):

                        command = f"""import deeplabcut; deeplabcut.label_frames(['{str(conf)}','{str(selected_vid)}'])"""
                        st.text("")
                        st.write(
                            "Please open an IPython console inside your antelope conda environment and copy and paste in the following command:"
                        )
                        st.code(command)
                        st.warning(
                            "This is a temporary measure while we figure out a bug in DeepLabCut"
                        )

                    st.text("")
                    st.info(
                        "Once you are happy with your labelled frames, use the button below to upload them to the database. You are required to do this before you can train your model on the cluster."
                    )
                    st.text("")
                    if st.button("Upload results"):

                        dlcup_pool = dlcup_thread_pool()
                        future = dlcup_pool.submit(upload_dlc, key)
                        st.session_state.upload_dlc_futures.append((future, name))
                        st.success("Upload in progress.")

                    if "upload_dlc_futures" in st.session_state:

                        if st.button("Check upload progress"):

                            st.write("Upload statuses:")

                            # initialise data
                            display_futures = []

                            # display job progress
                            for (
                                future,
                                query_name,
                            ) in st.session_state.upload_dlc_futures:

                                # compute statuses
                                if future.done():
                                    if future.exception():
                                        status = "upload error"
                                        print(future.exception())
                                    else:
                                        status = "upload success"
                                else:
                                    status = "upload in progress"

                                display_futures.append((query_name, status))

                            # make dataframe to display
                            df = pd.DataFrame(
                                display_futures, columns=["Query", "Status"]
                            )

                            # show dataframe
                            st.dataframe(df, hide_index=True)



from antelope.utils.os_utils import get_config, validate_config, get_config_path
import toml
import streamlit as st
import os


def show(tables, username):
    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:
        # just a widget to load the config and allow modifications plus saving
        # get the config path
        config = get_config()

        if config is None:
            config = {
                "deployment": {"deployment": "local"},
                "mysql": {"host": ""},
                "s3": {"host": ""},
                "folders": {},
                "analysis": {"folders": []},
                "multithreading": {"max_workers": 1},
                "computation": {
                    "host": "",
                    "basedir": "",
                    "antelope_data": "",
                },
            }

        new_config = {"deployment": {"deployment": "local"}}

        new_config["mysql"] = {}
        new_config["mysql"]["host"] = st.text_input(
            "Enter MySQL host", config["mysql"]["host"]
        )
        new_config["s3"] = {}
        new_config["s3"]["host"] = st.text_input("Enter S3 host", config["s3"]["host"])

        st.markdown("###### Add searchable folders")
        if "config_folders" not in st.session_state:
            st.session_state["config_folders"] = config["folders"]
        new_folders = {}
        for i, (name, path) in enumerate(st.session_state["config_folders"].items()):
            namecol, pathcol = st.columns([0.3, 0.7])
            with namecol:
                new_name = st.text_input("Enter name", name, key=f"name_{i}")
            with pathcol:
                new_path = st.text_input("Enter path", path, key=f"path_{i}")
            new_folders[new_name] = new_path
        st.session_state["config_folders"] = new_folders
        plus, minus = st.columns([0.05, 0.95])
        with plus:
            if st.button("\+"):
                st.session_state["config_folders"][""] = ""
                st.rerun()
        with minus:
            if st.button("\-"):
                st.session_state["config_folders"].popitem()
                st.rerun()

        st.markdown("###### Add analysis folders")
        if "config_analysis" not in st.session_state:
            st.session_state["config_analysis"] = config["analysis"]["folders"]
        new_analysis = []
        for i, path in enumerate(st.session_state["config_analysis"]):
            new_path = st.text_input("Enter path", path, key=f"analysis_{i}")
            new_analysis.append(new_path)
        st.session_state["config_analysis"] = new_analysis
        plus, minus = st.columns([0.05, 0.95])
        with plus:
            if st.button("\+", key="add_analysis"):
                st.session_state["config_analysis"].append("")
                st.rerun()
        with minus:
            if st.button("\-", key="remove_analysis"):
                st.session_state["config_analysis"].pop()
                st.rerun()

        new_config["multithreading"] = {}
        new_config["multithreading"]["max_workers"] = st.number_input(
            "Enter max worker threads", config["multithreading"]["max_workers"]
        )

        new_config["computation"] = {}
        new_config["computation"]["host"] = st.text_input(
            "Enter cluster host", config["computation"]["host"]
        )
        new_config["computation"]["basedir"] = st.text_input(
            "Enter antelope cluster installation location",
            config["computation"]["basedir"],
        )
        new_config["computation"]["antelope_data"] = st.text_input(
            "Enter antelope cluster data location",
            config["computation"]["antelope_data"],
        )

        if st.button("Save"):
            new_config["folders"] = st.session_state["config_folders"]
            new_config["analysis"] = {}
            new_config["analysis"]["folders"] = st.session_state["config_analysis"]

            if validate_config(new_config):
                config_path = get_config_path()
                os.makedirs(os.path.dirname(config_path), exist_ok=True)
                with open(config_path, "w") as f:
                    toml.dump(new_config, f)
                st.success("Config file saved")
            else:
                st.error("Invalid config file. Are all your paths correct?")



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.antelope_utils import (
    select_masks,
    recompute_masks,
    insert_masks,
    delete_masks,
)
from antelope.utils.analysis_utils import get_docstring
from antelope.utils.multithreading_utils import mask_thread_pool
import streamlit_antd_components as sac
import pandas as pd


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Compute Masks")

        st.divider()

        st.subheader("Select the sessions you want to recompute")

        tablename, world = dropdown_query_table(
            tables, {"Experimenter": tables["Experimenter"]}, username, headless=True
        )

        if "experimenter" in world.keys():
            # if no features in database raise warning
            if tablename == None:
                st.warning(
                    """You can't recompute any trials because you don't have the necessary upstream entries."""
                )
            elif len(tables["BehaviourRig"] & world) == 0:
                st.warning(
                    """You can't recompute any trials because you don't have the necessary upstream entries."""
                )
            elif username != world["experimenter"]:
                st.error("You can only edit your own data")
            else:

                st.text("")
                st.divider()

                sac.buttons(
                    [
                        sac.ButtonsItem(label="Change Mask Function"),
                        sac.ButtonsItem(label="Recompute Masks"),
                    ],
                    key="mask",
                    gap="md",
                    use_container_width=True,
                )

                if st.session_state["mask"] == "Change Mask Function":

                    # pull behaviour rig
                    rig_dicts = (
                        (tables["BehaviourRig"] & world)
                        .proj("behaviourrig_name")
                        .fetch(as_dict=True)
                    )
                    display_dict = {i["behaviourrig_name"]: i for i in rig_dicts}
                    rig = display_dict[
                        st.selectbox(
                            "Select the behaviour rig", list(display_dict.keys())
                        )
                    ]

                    # pull mask function
                    mask = (tables["MaskFunction"] & rig).fetch(format="frame")
                    mask["maskfunction_deleted"] = mask["maskfunction_deleted"].apply(
                        lambda x: True if x == "True" else False
                    )
                    mask = mask.reset_index()
                    del mask["experimenter"]
                    del mask["behaviourrig_id"]
                    if not mask.empty:
                        mask["mask_function"] = mask.apply(
                            lambda row: get_docstring(
                                row["mask_function"], row["maskfunction_name"]
                            ),
                            axis=1,
                        )
                    id = (
                        max((tables["MaskFunction"] & rig).fetch("mask_id"), default=0)
                        + 1
                    )

                    st.text("")
                    st.markdown("### Edit mask functions")
                    st.markdown("#### Current mask functions")
                    current_masks = st.data_editor(
                        mask,
                        hide_index=True,
                        use_container_width=True,
                        disabled=(
                            "mask_id",
                            "mask_name",
                            "mask_function",
                            "mask_description",
                        ),
                        column_config={
                            "mask_function": st.column_config.Column(width="small")
                        },
                    )

                    st.warning(
                        "Please note that the following performs a permanent delete on your masking functions"
                    )
                    st.text("")
                    if st.button("Delete selected masks"):
                        key = {
                            "experimenter": world["experimenter"],
                            "behaviourrig_id": rig["behaviourrig_id"],
                        }
                        if key["experimenter"] != username:
                            st.error("You can only delete your own masks")
                        else:
                            delete_masks(key, current_masks, tables)
                            st.success("Masks deleted")
                            st.rerun()

                    st.markdown("#### New mask functions")
                    edited_masks = select_masks(tables)
                    st.text("")
                    if st.button("Push new masks to database"):
                        if edited_masks:
                            masks = insert_masks(rig, edited_masks, id, tables)
                            tables["MaskFunction"].insert(masks)
                            st.success("Changes pushed to database")
                        else:
                            st.warning(
                                "No masks selected - nothing to push to database!"
                            )

                elif st.session_state["mask"] == "Recompute Masks":
                    st.text("")

                    # query sessions
                    table, world = dropdown_query_table(
                        tables,
                        {
                            "World": (tables["World"] - tables["Mask"])
                            & tables["MaskFunction"]
                        },
                        username,
                        headless=True,
                    )

                    if world:

                        # show how many sessions need recomputing
                        number_sessions = len(
                            (tables["World"] - tables["Mask"])
                            & tables["MaskFunction"]
                            & world
                        )
                        st.text("")
                        st.write(f"Computing masks for {number_sessions} sessions")
                        st.text("")

                        if st.button("Recompute masks"):

                            # retrieve thread pool
                            thread_pool = mask_thread_pool()

                            # submit job to thread pool
                            future = thread_pool.submit(recompute_masks, world)

                            # append future to session state
                            st.session_state.mask_futures.append((future, world))
                            st.text("")
                            st.success("Computation in progress!")

                        st.info(
                            "Note that computing many masks can take a while. This will occur in a separate thread so you can still use Antelope while the computation is occurring, and can use the button below to check your computation status."
                        )

                        st.text("")

                        if "mask_futures" in st.session_state:

                            if st.button("Check computation progress"):

                                st.write("Computation statuses:")

                                # initialise data
                                display_futures = []

                                # compute job statuses
                                for future, world in st.session_state.mask_futures:

                                    # compute statuses
                                    if future.done():
                                        if future.exception():
                                            status = "computation error"
                                        else:
                                            status = "computation success"
                                    else:
                                        status = "computation in progress"

                                    # primary keys for display
                                    keys = {
                                        key: val
                                        for key, val in world.items()
                                        if key in tables[tablename].primary_key
                                    }
                                    display = "-".join([str(i) for i in keys.values()])

                                    display_futures.append((display, status))

                                # make dataframe to display
                                df = pd.DataFrame(
                                    display_futures, columns=["Primary Key", "Status"]
                                )

                                # show dataframe
                                st.dataframe(df, hide_index=True)

                    else:

                        st.info("No masks to recompute")



import streamlit as st
from antelope.utils.external_utils import schedule_train_dlc, train_dlc_progress
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.antelope_utils import display_dlc_images
import streamlit_antd_components as sac


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Train your DeepLabCut model")

        st.divider()
        sac.buttons(
            [
                sac.ButtonsItem(label="Schedule Model Training"),
                sac.ButtonsItem(label="Evaluate Model"),
            ],
            key="dlcmode",
            gap="md",
            use_container_width=True,
        )

        if st.session_state.dlcmode == "Schedule Model Training":
            st.subheader("Select the model you want to train")

            # Select the feature to add
            tablename, key = dropdown_query_table(
                tables,
                {
                    "LabelledFrames": tables["LabelledFrames"] - tables["DLCModel"]
                    & {"labelledframes_in_compute": "False"}
                },
                username,
                headless=True,
            )
            if key:
                key["behaviourrig_deleted"] = "False"

            if key is None:
                st.error(
                    """You can't train a DeepLabCut model yet as you haven't labelled any frames or all possible models are already trained."""
                )
                st.warning("Please go to the insert tab and label some frames.")

            else:

                st.divider()

                # schedule computation section
                st.subheader("Schedule computation")

                # ask for user password for scheduling the job
                password = st.text_input(
                    "Please enter your cluster password", type="password"
                )

                st.text("")
                if st.button("Schedule model training"):

                    if key["experimenter"] != username:

                        st.error("You can only train models on your own data")

                    else:

                        try:

                            num_videos = len(
                                tables["World"] * tables["Video"]
                                & key
                                & {
                                    "dlc_training": "True",
                                    "world_deleted": "False",
                                    "video_deleted": "False",
                                }
                            )

                            # send job
                            schedule_train_dlc(key, num_videos, password)

                            # success message
                            st.success("Job sent to cluster!")
                            st.info(
                                "You will receive an email once your job is completed"
                            )

                        except Exception:

                            st.error("Error submitting job to cluster")

                st.text("")

                # if there are any downlaods this session
                if "train_dlc_jobs" in st.session_state:

                    # button which shows spikesort statuses
                    if st.button("Check model training progress"):

                        train_dlc_progress()

        if st.session_state.dlcmode == "Evaluate Model":

            st.subheader("Select the model you want to evaluate")

            # Select the feature to add
            tablename, key = dropdown_query_table(
                tables, {"DLCModel": tables["DLCModel"]}, username, headless=True
            )

            if key is None:
                st.error(
                    """You can't evaluate a DeepLabCut model yet as you haven't trained any models."""
                )
                return

            # fetch results and display
            st.divider()
            st.markdown("### Evaluation metrics")
            metrics = (tables["DLCModel"] & key).fetch1("evaluation_metrics")
            st.dataframe(metrics, use_container_width=True)

            # fetch images and display
            st.divider()
            if st.button("Show evaluated images"):

                st.session_state["dlc_inference_folder"] = display_dlc_images(
                    tables["DLCModel"], key
                )

            # display images
            if "dlc_inference_folder" in st.session_state:
                folder = st.session_state["dlc_inference_folder"]

                images = list(folder.glob("*.png"))

                test_images = [i for i in images if "Test" in i.name]
                train_images = [i for i in images if "Train" in i.name]

                image_dict = {
                    f"Test {i+1}": str(test_images[i].resolve())
                    for i in range(len(test_images))
                }
                image_dict.update(
                    {
                        f"Train {i+1}": str(train_images[i].resolve())
                        for i in range(len(train_images))
                    }
                )

                image = image_dict[
                    st.selectbox("Select image to display", list(image_dict.keys()))
                ]

                st.image(image)



import streamlit as st
import streamlit_antd_components as sac
from antelope.utils import analysis_utils
from antelope.utils.streamlit_utils import (
    dropdown_query_table,
    enter_args,
    display_analysis,
)
from antelope.utils.external_utils import schedule_analysis, analysis_progress
from antelope.utils.analysis_utils import reload_analysis


def reset_state():
    st.session_state.result = None


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Analysis")
        st.subheader("Run analysis on the database")

        def reset_state():
            st.session_state.result = None

        st.divider()
        if not hasattr(st.session_state, "result"):
            reset_state()

        # load analysis functions
        analysis_functions = analysis_utils.import_analysis(
            tables["Experimenter"].connection, tables
        )

        # get user to select the function they want
        st.markdown("#### Select function")
        folder = st.selectbox(
            "Select analysis folder",
            list(analysis_functions.keys()),
            on_change=reset_state,
        )
        func_dict = {f.name: f for f in analysis_functions[folder] if f.hidden == False}
        name = st.selectbox(
            "Select analysis function",
            list(func_dict.keys()),
            on_change=reset_state,
        )
        function = func_dict[name]

        # display docstring
        st.text("")
        st.markdown(function.__doc__)
        if st.button("Reload functions"):
            analysis_functions = reload_analysis(
                tables["Experimenter"].connection, tables
            )
            st.rerun()
        st.divider()

        # get user to select the restriction they want to apply
        st.markdown("#### Enter restriction")
        if hasattr(function, "key"):
            key = function.key
        else:
            key = {}
        if isinstance(function.query, str):
            table = tables[function.query].proj()
        elif isinstance(function.query, list):
            table = tables[function.query[0]]
            for q in function.query[1:]:
                table = table * tables[q].proj()
        _, restriction = dropdown_query_table(
            tables, {"table": table & key}, username, headless=True
        )

        if _ == None:
            st.error("No data available to run analysis on.")

        else:
            # get the user to input the arguments
            if hasattr(function, "args") and function.args:
                st.divider()
                st.markdown("#### Enter arguments")
                args = enter_args(function)
            else:
                args = {}

            # run the function
            st.divider()
            st.markdown("#### Select mode")
            sac.buttons(
                [sac.ButtonsItem(label="Local"), sac.ButtonsItem(label="Cluster")],
                key="mode",
                use_container_width=True,
            )

            if st.session_state.mode == "Local":
                st.text("")
                result = None
                if st.button("Run"):

                    @st.cache_data(ttl=600)
                    def cache_fct(name, restriction, **args):
                        fct = func_dict[name]
                        try:
                            return fct(restriction, **args)
                        except Exception as e:
                            st.error(
                                f"""
                            Your function has an error.\n
                            Error message:\n
                            {e}
                            """
                            )

                    result = cache_fct(name, restriction, **args)
                    st.session_state.result = result

                st.divider()
                if hasattr(st.session_state, "result"):
                    display_analysis(
                        st.session_state.result, function.returns, table.primary_key
                    )

            if st.session_state.mode == "Cluster":
                st.text("")

                savepath = st.text_input("Enter the path to save the results to")
                numcpus = st.number_input(
                    "Enter the number of cpus to use",
                    min_value=1,
                    max_value=64,
                    value=1,
                )
                time = st.number_input(
                    "Enter the time to run the job for (minutes)",
                    min_value=10,
                    max_value=24 * 60,
                    value=60,
                )
                password = st.text_input(
                    "Please enter your cluster password", type="password"
                )

                st.text("")
                if st.button("Schedule analysis job"):

                    try:
                        schedule_analysis(
                            function,
                            restriction,
                            savepath,
                            numcpus,
                            time,
                            password,
                            args,
                        )
                    except Exception as e:
                        print(e)
                        st.error(
                            "Job scheduling failed. Please check your inputs and try again."
                        )
                    else:
                        st.text("")
                        st.success("Job sent to cluster.")

                # if there are any downlaods this session
                if "analysis_jobs" in st.session_state:

                    # button which shows spikesort statuses
                    st.text("")
                    if st.button("Check analysis progress"):

                        analysis_progress()



import numpy as np
from antelope import antelope_analysis
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d


@antelope_analysis
class Greeting:
    """
    This is Antelope's hello world function.
    """

    name = "greeting"
    query = "Experimenter"
    returns = {"greeting": str}
    args = {"excited": bool}

    def run(key, excited=True):
        full_name = (Experimenter & key).fetch1("full_name")
        if excited:
            return f"Hello, {full_name}!"
        else:
            return f"Hello, {full_name}."


@antelope_analysis
class CountExperiments:
    """
    This is a slightly more complex example showing how we can aggregate over another table and rename variables within the function.
    It's worth noting that when you aggregate, the argument passed to the function will always be a list.
    """

    name = "count_experiments"
    query = "Experimenter"
    data = "Experiment"
    returns = {"count": int}

    def run(key):
        length = len(Experiment & key)
        return length


@antelope_analysis
class GreetingWithCount:
    """
    This example shows how we can build on top of other functions and use multiple attributes, both within the same table and from different tables.
    To do so, we need to define the other functions we want to run in the `inherits` attribute, and pass them as inputs to the function.
    These inner functions can then be run with any restriction - although the typical use case is to use a primary key.
    """

    name = "greeting_with_count"
    query = "Experimenter"
    returns = {"response": str}
    calls = ["greeting", "count_experiments"]

    def run(key):
        greet = greeting(key)["greeting"]
        num_experiments = count_experiments(key)["count"]
        institution = (Experimenter & key).fetch1("institution")
        response = (
            f"{greet} You have run {num_experiments} experiments at {institution}."
        )
        return response


@antelope_analysis
class FirstExperimentName:
    """
    This example shows how we can use a restriction to filter the data within the function.

    Restrictions can of course be passed when running the function, but are useful at this level
    to define when the function doesn't apply to certain attributes, or more commonly, to define
    different subsets of aggregated attributes as different inputs to the function.

    Note, you should always handle the case where the function input is an empty list.
    """

    name = "first_experiment_name"
    query = "Experimenter"
    returns = {"response": str}

    def run(key):
        experiment_name = (Experiment & key).fetch("experiment_name", limit=1)
        if len(experiment_name) == 0:
            return "You have not run any experiments."
        elif len(experiment_name) == 1:
            return f"The first experiment you ran was called {experiment_name[0]}."
        else:
            raise ValueError("This error should never get raised.")


@antelope_analysis
class ExampleFigure:
    """
    Example of a function that returns a matplotlib figure.
    """

    name = "example_figure"
    query = "Experimenter"
    returns = {"figure": plt.Figure}
    args = {"size": int}

    def run(key, size=10):
        full_name = (Experimenter & key).fetch1("full_name")
        fig, ax = plt.subplots()
        ax.text(0.5, 0.5, f"Hello\n{full_name}!", fontsize=size, ha="center")
        ax.axis("off")
        return fig


@antelope_analysis
class Sta:
    """
    The spike-triggered average for an analog event.

    This example shows how for some functions, it makes sense to define the function as running on the join of two tables.
    """

    name = "sta"
    query = ["SpikeTrain", "AnalogEvents"]
    returns = {"Spike-triggered average": np.ndarray, "Timestamps (s)": np.ndarray}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        spiketrain = (SpikeTrain & key).fetch1("spiketrain")
        data, timestamps = (AnalogEvents.proj("data", "timestamps") & key).fetch1(
            "data", "timestamps"
        )

        # interpolate the event data
        event_func = interp1d(timestamps, data, fill_value=0, bounds_error=False)

        # create window timestamps
        step = 1 / sample_rate
        start_time = -(window_size // step) * step
        window_timestamps = np.arange(start_time, 0, step)

        # create matrix of window times for each spike - shape (n_spikes, window_samples)
        sta_times = spiketrain[:, None] + window_timestamps

        # get the event values in each window
        sta_values = event_func(sta_times)

        # average over all spikes
        sta = np.mean(sta_values, axis=0)

        return sta, window_timestamps



import datajoint as dj
from antelope.schemas import session, ephys, behaviour
from antelope.connection.connect import dbconnect
from PIL import Image, ImageOps, ImageFilter


# import schemas
conn = dbconnect()
session_tables, session_schema = session.schema(conn)
ephys_tables, ephys_schema = ephys.schema(conn)
behaviour_tables, behaviour_schema = behaviour.schema(conn)

for key, DigitalEvents in session_tables.items():
    globals()[key] = DigitalEvents

for key, DigitalEvents in ephys_tables.items():
    globals()[key] = DigitalEvents

for key, DigitalEvents in behaviour_tables.items():
    globals()[key] = DigitalEvents

# plot various diagrams
(dj.Diagram(session_schema)).save("resources/session.png")

(dj.Diagram(session_schema) + dj.Diagram(ephys_schema)).save("resources/ephys.png")

(dj.Diagram(session_schema) + dj.Diagram(behaviour_schema)).save(
    "resources/behaviour.png"
)


# function to remove white background from the images and reshape
def remove_white_background(
    image_path, output_path, target_size, tolerance, radius, target_colour
):

    # Open the image
    img = Image.open(image_path)

    # Convert to RGBA mode if not already
    img = img.convert("RGBA")

    # Get the image data
    img_data = img.getdata()

    # Remove white pixels
    new_data = []
    for item in img_data:
        # Check if the pixel is white or near white within the defined tolerance
        if all(abs(channel - 255) <= tolerance for channel in item[:3]):
            # Set alpha to 0 for white pixels
            new_data.append(target_colour + (item[3],))
        else:
            new_data.append(item)

    # Update image with new data
    img.putdata(new_data)
    if radius != None:

        # Create a mask to identify boundary pixels
        boundary_mask = Image.new("L", img.size, 0)
        for x in range(img.width):
            for y in range(img.height):
                pixel = img.getpixel((x, y))
                if pixel[3] == 0:  # Transparent pixel
                    for dx in range(-1, 2):
                        for dy in range(-1, 2):
                            nx, ny = x + dx, y + dy
                            if 0 <= nx < img.width and 0 <= ny < img.height:
                                neighbor_pixel = img.getpixel((nx, ny))
                                if neighbor_pixel[3] != 0:  # Non-transparent neighbor
                                    boundary_mask.putpixel((x, y), 255)
                                    break

        # Apply Gaussian blur to boundary pixels
        blurred_boundary_mask = boundary_mask.filter(
            ImageFilter.GaussianBlur(radius=radius)
        )  # Adjust radius as needed

        # Apply blur only to boundary pixels
        blurred_img = img.copy()
        for x in range(img.width):
            for y in range(img.height):
                if blurred_boundary_mask.getpixel((x, y)) > 0:
                    blurred_img.putpixel((x, y), (255, 255, 255, 0))  # Set alpha to 0

        img = blurred_img.copy()

    # Pad the image to the target size
    img = img.resize(target_size)
    padded_img = ImageOps.expand(img, border=(300, 200), fill=target_colour)

    # Save the new image
    padded_img.save(output_path, "PNG")


# remove white background from the images
target_size = (400, 600)
tolerance = 0
target_colour = (29, 34, 41)
remove_white_background(
    "resources/session.png",
    "resources/session.png",
    (200, 300),
    tolerance,
    radius=None,
    target_colour=target_colour,
)
remove_white_background(
    "resources/ephys.png",
    "resources/ephys.png",
    target_size,
    tolerance,
    radius=None,
    target_colour=target_colour,
)
remove_white_background(
    "resources/behaviour.png",
    "resources/behaviour.png",
    target_size,
    tolerance,
    radius=None,
    target_colour=target_colour,
)



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.visualisation_utils import plot_units, plot_isi, unit_stats
import streamlit_antd_components as sac


def show(username, tables):

    st.title("Visualisation")
    st.subheader("Spikesorting unit waveforms and statistics")
    st.divider()

    col1, col2 = st.columns([0.4, 0.6], gap="large")

    with col1:

        st.markdown("**Select the spikesorting you want to visualise**")

        # get user to interactively insert table attributes
        tablename, spikesorting_dict = dropdown_query_table(
            tables, {"SpikeSorting": tables["SpikeSorting"]}, username, headless=True
        )

        if spikesorting_dict is None:
            st.warning("You currently have no spikesorted data to visualise")

        elif "sortingparams_id" in spikesorting_dict.keys():

            # first query the probes
            probes = (tables["Probe"] & spikesorting_dict).fetch(as_dict=True)

            # initialise selected units
            selected_units = []

            st.text("")
            st.text("Select the units you want to visualise")

            # loop through probes and get user to select units
            for probe in probes:

                # get units for this probe
                units = (tables["Unit"] & spikesorting_dict & probe).fetch(as_dict=True)

                # make a dict mapping unit names to unit dicts
                unit_dict = {}
                for unit in units:
                    unit_dict[str(unit["unit_id"])] = unit

                # get user to select the units they want to visualise
                selected_units += [
                    unit_dict[i]
                    for i in sac.checkbox(
                        items=list(unit_dict.keys()),
                        label=f"""Probe {probe['probe_id']}""",
                        index=list(range(len(unit_dict))),
                        check_all="Select all",
                    )
                ]

            if st.button("Plot"):

                # generate plots
                st.session_state["waveforms_vis"] = plot_units(
                    spikesorting_dict, selected_units
                )
                st.session_state["isi_vis"] = plot_isi(
                    spikesorting_dict, selected_units
                )
                st.session_state["unit_stats"] = unit_stats(
                    spikesorting_dict, selected_units
                )

            with col2:

                if "waveforms_vis" in st.session_state.keys():

                    wav_figures = st.session_state["waveforms_vis"]
                    isi_figures = st.session_state["isi_vis"]
                    stats = st.session_state["unit_stats"]

                    st.markdown("**Select a probe**")

                    # button displays different probes
                    probe = sac.buttons(
                        [sac.ButtonsItem(label=probe) for probe in wav_figures.keys()],
                        format_func=lambda x: f"Probe {x}",
                        use_container_width=True,
                    )

                    mode = sac.buttons(
                        [
                            sac.ButtonsItem(label="Waveforms"),
                            sac.ButtonsItem(label="ISI histograms"),
                            sac.ButtonsItem(label="Statistics"),
                        ],
                        use_container_width=True,
                    )

                    # display the selected probe
                    if mode == "Waveforms":
                        st.plotly_chart(wav_figures[probe], use_container_width=True)
                    elif mode == "ISI histograms":
                        st.plotly_chart(isi_figures[probe], use_container_width=True)
                    elif mode == "Statistics":
                        st.dataframe(
                            stats[probe], hide_index=True, use_container_width=True
                        )



"""
Utility functions for datajoint table manipulations

Author: Rory Bedfod
Date: 07/07/23
"""

import pandas as pd
import datajoint as dj
import numpy as np
import pathlib
from pathlib import Path
import os
import sys
import shutil
from io import StringIO
import json
import datetime
from antelope.connection import connect, import_schemas
import tempfile


def delete_status(tablename, tables, query):
    """
    Function returns the delete status of a table
    Options include 'temp', 'perm', 'none', 'custom-behave'
    """

    from antelope.utils.antelope_utils import check_animals

    parent_tables = [
        i for i in tables.values() if i.full_table_name in tables[tablename].parents()
    ]

    table = tables[tablename]

    no_delete = ["Feature"]

    # can't delete part tables for concurrency
    if tablename in no_delete:
        return "none"

    # custom behaviour for self and world tables
    elif (
        tablename == "Self"
        and len(check_animals((tables["BehaviourRig"] & query).fetch1("rig_json"))) == 1
    ):
        return "custom-behave"

    # if there's a spare key, can temp delete
    elif check_spare_key(table, parent_tables):
        return "temp"

    # if there's no spare key, permanent delete
    else:
        return "perm"


def check_spare_key(table, parent_tables):
    """
    Function checks if there's a spare key in the table primary keys, not belonging to any parent
    """

    # construct join query of all parents
    counter = 0
    for parent in parent_tables:
        if counter == 0:
            parent_join = parent.proj()
            counter += 1
        else:
            parent_join = parent_join * parent.proj()

    # check if there's a key not defined in parents
    spare_key = False
    for key in table.primary_key:
        if key not in parent_join.heading.attributes.keys():
            spare_key = True
            break

    return spare_key


def insertable_tables(tables, subtables, username):
    """
    Returns the subdictionary of all tables the user can insert into
    """
    available_tables = dict()

    for tablename, table in subtables.items():

        # compute all parents of this table
        parent_tables = [
            i for i in tables.values() if i.full_table_name in table.parents()
        ]

        spare_key = check_spare_key(table, parent_tables)

        # if there's a spare key, and all parent have undeleted entries
        if spare_key:

            if len(parent_tables) > 0 and all(
                [len(i & {"experimenter": username}) > 0 for i in parent_tables]
            ):

                available_tables[tablename] = table

        # if there's no spare key, and there are remaining combinations of parent keys
        else:

            counter = 0
            for parent in parent_tables:
                if counter == 0:
                    parent_join = parent.proj()
                    counter += 1
                else:
                    parent_join = parent_join * parent.proj()

            # if there are any entries in the join query not already in the table
            if len(parent_join - table & {"experimenter": username}) > 0:
                available_tables[tablename] = table

    # can't insert into experimenter table
    if "Experimenter" in available_tables.keys():
        del available_tables["Experimenter"]

    return available_tables


def searchable_tables(tables, delete_mode):
    """
    Returns the subdictionary of tables the user is able to search
    """

    available_tables = dict()

    for tablename, table in tables.items():
        if table & delete_restriction(table, delete_mode):
            available_tables[tablename] = table

    return available_tables


def form_query(querystring, tables):
    """
    Forms a valid datajoint query object from an input string
    Supports all the tables, operators, and restrictions
    Should return errors on any misformed queries (provides some protection against malignant SQL)
    Inputs: querystring: string representing datajoint query
    Returns: query: a valid datajoint query object
    """

    # first, move variables into local namespace
    for tablename in tables.keys():
        locals()[tablename] = tables[tablename]

    # convert text to actual query object
    query = eval(querystring)

    return query


def parent_primaries(table, tables, primary_only=False):
    """
    Finds the parent tables referenced by all foreign keys for a given table

    Inputs: table: the table to apply the operation to
            tables: a dictionary of all tables in the database
            primary_only: if True, only return for primary keys of the table

    Returns: a dictionary of attribute:table mappings for all foreign keys
    """

    # calculate all parents of table
    full_names = {val.full_table_name: key for key, val in tables.items()}
    if type(table) == dj.expression.QueryExpression:

        join_tables = table.support
        parents = []
        for join_table in join_tables:
            parents += join_table.parents()
        parents = list(set(parents))
    else:
        parents = table.parents()

    # dictionary with parent primary keys mapping to the parent tables for querying
    parent_tables = {}
    for par_table in tables.values():
        for parent in parents:
            if par_table.full_table_name == parent:
                for primary_key in par_table.primary_key:
                    parent_tables[primary_key] = par_table

    if primary_only:
        primary_keys = table.primary_key
        parent_tables = {
            key: val for key, val in parent_tables.items() if key in primary_keys
        }

    return parent_tables


def ancestor_primaries(table, tables):
    """
    Finds the earliest ancestor tables referenced by all foreign keys for a given table

    Inputs: table: the table to apply the operation to
            tables: a dictionary of all tables in the database

    Returns: a dictionary of attribute:table mappings for all foreign keys
    """

    # first return all parent tables for all foreign keys
    parent_tables = parent_primaries(table, tables)
    next_tables = parent_tables.copy()
    result = {}

    # loop as long as there are still upstream tables
    while next_tables:

        # now for all referenced tables, calculate all upstream tables
        next_tables = {}
        for key, tab in parent_tables.items():
            next_tables.update(parent_primaries(tab, tables, primary_only=True))

        # calculate tables that have finished
        result.update(
            {
                key: val
                for key, val in parent_tables.items()
                if key not in next_tables.keys()
            }
        )
        parent_tables = {
            key: val for key, val in next_tables.items() if key not in result.keys()
        }
        next_tables = dict(parent_tables)

    return result


def query_without_external(query, mode="Sequential Filter"):
    """
    Queries the database without downloading external attributes or displaying jsons

    Inputs: table: table to be queried
            restriction: optional dict of restrictions

    Outputs: dataframe of query with up to 30 items (with placeholders for external and json)
             length of full query
    """

    # calculate any external attributes to be removed (so it doesn't download)
    # also calculate any jsons to be removed (since they look messy)
    projection = []
    external = []
    jsons = []
    blobs = []
    for key, val in query.heading.attributes.items():
        if val.is_attachment:
            external.append(key)
        elif val.json:
            jsons.append(key)
        elif val.is_blob:
            blobs.append(key)
        else:
            projection.append(key)

    # remove external attributes and deleted
    query = query.proj(*projection)

    # fetch query
    df = query.fetch(format="frame", limit=30)

    # reset index (so we can hide it later)
    df = df.reset_index()

    # replace json entries
    for i in jsons:
        df[i] = len(df.index) * ["json"]

    # replace blobs entries
    for i in blobs:
        df[i] = len(df.index) * ["blob"]

    # add blank external entries for display
    for i in external:
        df[i] = len(df.index) * ["external"]

    return df, len(query)


def download_data(querystring, download_path, mode="Sequential Filter"):
    """
    Downloads query as a numpy array as well as external data.
    Inputs: querystring: the query in string form (needs to be serialisble - queries themselves include connections)
            download_path: path to download data to
    """

    # function runs in different process so needs to establish its own connection
    conn_download = connect.connect()
    tables_download = import_schemas.schema(conn_download)

    # form query
    query = form_query(querystring, tables_download)

    # append keys we want to keep
    projection = []
    for key in query.heading.attributes.keys():
        if mode == "Sequential Filter":
            continue
        elif key == f"""{query.table_name.replace('_','')}_in_compute""":
            continue
        else:
            projection.append(key)

    # remove deleted attribute from query
    query = query.proj(*projection)

    # fetch results of query
    results = query.fetch(download_path=download_path)

    # loop through tables attributes to change external paths
    keys = list(query.heading.attributes.keys())
    for key in keys:

        # if attribute is an external file
        if query.heading.attributes[key].is_attachment:

            # convert to dataframe
            df = pd.DataFrame(results)

            # loop and change path to current directory
            for index, row in df.iterrows():

                # load path
                path = pathlib.Path(row[key])

                # update to just the filename
                row[key] = path.name

            # convert back to recarray
            results = df.to_records()

    # save array to disk
    np.save(str(download_path), results)


def check_session(tables, insert_dict):
    """
    function checks if directory has correct extension
    """

    # first, check the directory structure is correct
    # load config file
    resources = Path(os.path.abspath(__file__)).parent.parent / "resources"
    with open(resources / "ephys_extensions.json") as f:
        extensions = json.load(f)

        # query equipment type
        ephys_acquisition = insert_dict["ephys_acquisition"]

        # check if ephys acquisition is none
        if ephys_acquisition == "none":
            return False

        #  get list of extensions
        equip_ext = extensions[ephys_acquisition]
        checkls = []

        # check all files exist
        for ext in equip_ext:
            status = len(list(insert_dict["recording"].glob(f"*.{ext}"))) > 0
            checkls.append(status)

        # return true if all files exist
        return all(checkls)


def upload(tablename, insert_dict, mode="insert"):
    """
    Uploads data to database including external data
    Inputs: table: the table to insert to
            insert_dict: the content of the upload as a dictionary
    """

    # function runs in different process so needs to establish its own connection
    conn_upload = connect.connect()
    tables_upload = import_schemas.schema_admin(conn_upload)

    # get table
    table_upload = tables_upload[tablename]

    # make dict mapping full_table_names to display table full_table_names
    full_names = {val.full_table_name: key for key, val in tables_upload.items()}

    # if it's a session, we need to zip the raw data first and update the path
    if tablename == "Recording":

        # extract dirpath
        dirpath = insert_dict["recording"]

        # create zipfile name from primary key
        zipname = "-".join([str(list(insert_dict.values())[i]) for i in range(4)])
        zipfile = Path(tempfile.gettempdir()) / "antelope_raw" / pathlib.Path(zipname)

        # create zipfile
        shutil.make_archive(zipfile, "zip", dirpath)

        # update dirpath
        insert_dict["recording"] = str(zipfile) + ".zip"

    # start insert transaction
    with table_upload.connection.transaction:

        # insert data
        if mode == "insert":
            table_upload.insert1(insert_dict)

        elif mode == "update":
            table_upload.update1(insert_dict)

        # after insertion, need to check the delete status of the upstream tables
        # as these could get safe deleted during the insert
        for parentname in table_upload.parents():

            # retrieve datajoint table from full name
            parent = tables_upload[full_names[parentname]]

            # check parent delete status
            if full_names[parentname] == "Experimenter":
                delete_status = "False"
            else:
                delete_status = (parent & insert_dict).fetch1(
                    f"""{parent.table_name.replace('_','')}_deleted"""
                )

            # if any parents are deleted
            if delete_status == "True":

                # raise error so thread status checker works and transaction rolls back
                raise Exception("A parent table has been deleted during the upload.")


def safe_delete(querystring):
    """
    Safely deletes data from the database (just changes the 'delete' attribute to True)
    These deletes cascade
    Inputs: tablename: the table to delete from
            restriction: the key to delete
    """
    # function runs in different process so needs to establish its own connection
    conn_del = connect.connect()
    tables_del = import_schemas.schema_admin(conn_del)

    # form query
    query = form_query(querystring, tables_del)

    update_dict = {}

    # make dict mapping full_table_names to display table full_table_names
    full_names = {val.full_table_name: key for key, val in tables_del.items()}

    # all updates in transaction
    with conn_del.transaction:

        # loop through all tables to be modified
        for tablename in query.descendants():

            # retrieve datajoint table from full name
            table = tables_del[full_names[tablename]]

            # main query
            full_query = table & (
                query.proj()
                & {f"""{query.table_name.replace('_','')}_deleted""": "False"}
            )

            # fetch primary keys of rows to be deleted
            data = full_query.proj().fetch(as_dict=True)

            update_dict[tablename] = data

        # now loop again to perform modifications
        for tablename, data in update_dict.items():

            # retrieve datajoint table from full name
            table = tables_del[full_names[tablename]]

            for i in data:

                table.update1(
                    {**i, f"""{table.table_name.replace('_','')}_deleted""": "True"}
                )


def restore(querystring):
    """
    Restores deleted data from the database (just changes the 'delete' attribute to False)
    These restores cascade
    Inputs: querystring: the query to restore from
    """
    # function runs in different process so needs to establish its own connection
    conn_res = connect.connect()
    tables_res = import_schemas.schema_admin(conn_res)

    # form query
    query = form_query(querystring, tables_res)

    update_dict = {}

    # make dict mapping full_table_names to display table full_table_names
    full_names = {val.full_table_name: key for key, val in tables_res.items()}

    # all updates in transaction
    with conn_res.transaction:

        # loop through all tables to be modified
        for tablename in query.descendants():

            # retrieve datajoint table from full name
            table = tables_res[full_names[tablename]]

            # construct query to delete
            tmp_query = (
                table
                & query.proj()
                & {f"""{table.table_name.replace('_','')}_deleted""": "True"}
            )

            # additionally, need to check the parents are either descendants of the query, or are undeleted
            for parentname in table.parents():
                if parentname not in query.descendants():
                    parent = tables_res[full_names[parentname]]

                    tmp_query = tmp_query & (
                        parent
                        & {f"""{parent.table_name.replace('_','')}_deleted""": "False"}
                    )

            # fetch primary keys of rows to be deleted
            data = tmp_query.proj().fetch(as_dict=True)

            update_dict[tablename] = data

        for tablename, data in update_dict.items():

            table = tables_res[full_names[tablename]]

            for i in data:

                table.update1(
                    {**i, f"""{table.table_name.replace('_','')}_deleted""": "False"}
                )


def release_computation(tablename, restriction):
    """
    Releases data from computation(just changes the 'in_compute' attribute to False)
    Inputs: querystring: the query to release from
    """
    # function runs in different process so needs to establish its own connection
    conn_rel = connect.connect()
    tables_rel = import_schemas.schema_admin(conn_rel)

    # form datajoint querystring
    querystring = f"{tablename} & {str(restriction)}"

    # form query
    query = form_query(querystring, tables_rel)

    table = tables_rel[tablename]

    # all updates in transaction
    with conn_rel.transaction:

        if tablename == "SpikeSorting":
            # get data to update
            data = query.proj("manually_curated").fetch(as_dict=True)

            # loop through data and update or delete
            for i in data:
                if i["manually_curated"] == "True":
                    table.update1(
                        {
                            **i,
                            f"""{table.table_name.replace('_','')}_in_compute""": "False",
                        }
                    )
                elif i["manually_curated"] == "False":
                    with SimulateEnterContext():
                        (table & i).delete(safemode=False)

        elif tablename == "LabelledFrames":
            data = query.proj().fetch(as_dict=True)
            for i in data:
                table.update1(
                    {**i, f"""{table.table_name.replace('_','')}_in_compute""": "False"}
                )


def true_delete(query):
    """
    Deletes data from the database
    Inputs: query: the query to delete
    """

    # delete data
    query.delete(safemode=False)


def show_deletes(tables, query):
    """
    Function to show all the deletes that cascade
    """

    # make dict mapping full_table_names to display table full_table_names
    full_names = {val.full_table_name: key for key, val in tables.items()}

    # initialise dataframe with number of deletes from each table
    descendant_dict = {"Table": [], "Number entries to be deleted": []}

    # cycle through descendants calculating number that will get deleted
    for descendant in query.descendants():

        table = tables[full_names[descendant]]

        # main query
        full_query = table & query.proj()

        number_entries = len(full_query)
        if number_entries > 0:
            descendant_dict["Table"].append(full_names[descendant])
            descendant_dict["Number entries to be deleted"].append(number_entries)

    return descendant_dict


def show_restores(tables, query):
    """
    Function to show all the restores that cascade
    """

    # make dict mapping full_table_names to display table full_table_names
    full_names = {val.full_table_name: key for key, val in tables.items()}

    # initialise dataframe with number of deletes from each table
    descendant_dict = {"Table": [], "Number entries to be restored": []}

    # cycle through descendants calculating number that will get deleted
    for descendant in query.descendants():

        table = tables[full_names[descendant]]

        # want entries that are deleted and join the query
        tmp_query = table * query.proj() & {
            f"""{table.table_name.replace('_','')}_deleted""": "True"
        }

        # additionally, need to check the parents are either descendants of the query, or are undeleted
        if query.table_name not in descendant:  # skip if parent table
            for parentname in table.parents():

                if parentname not in query.descendants():

                    parent = tables[full_names[parentname]]

                    tmp_query = tmp_query & parent

        number_entries = len(tmp_query)
        if number_entries > 0:
            descendant_dict["Table"].append(full_names[descendant])
            descendant_dict["Number entries to be restored"].append(number_entries)

    return descendant_dict


def show_true_deletes(tables, query):
    """
    Function to show all the restores that cascade
    """

    # make dict mapping full_table_names to display table full_table_names
    full_names = {val.full_table_name: key for key, val in tables.items()}

    # initialise dataframe with number of deletes from each table
    descendant_dict = {"Table": [], "Number entries to be restored": []}

    # cycle through descendants calculating number that will get deleted
    for descendant in query.descendants():

        table = tables[full_names[descendant]]

        # want entries that are deleted and join the query
        tmp_query = table * query.proj() & {
            f"""{table.table_name.replace('_','')}_deleted""": "True"
        }

        number_entries = len(tmp_query)
        if number_entries > 0:
            descendant_dict["Table"].append(full_names[descendant])
            descendant_dict["Number entries to be restored"].append(number_entries)

    return descendant_dict


def query_to_str(query):
    """
    Function makes an easily readable string from a datajoint query
    Used for file names and identifying queries
    """

    # extract SQL
    string = query.make_sql()

    # take table name
    string = string.split("FROM")[1]
    string = string.split(".")[1]
    string = string.split(" ")[0]

    # clean tablename
    string = (
        string.replace("`", "")
        .replace("#", "")
        .replace(".", "")
        .replace("_", "")
        .strip()
    )

    # final name
    name = (
        "antelope-"
        + string
        + "--"
        + datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
    )

    return name


def delete_restriction(query, delete_mode="False"):

    cols = list((query & False).fetch(format="frame").reset_index().columns)
    delete_dict = {}
    for col in cols:
        if "deleted" in col:
            delete_dict[col] = delete_mode
    return delete_dict


def delete_column(query):

    cols = list((query & False).fetch(format="frame").reset_index().columns)
    to_delete = []
    not_delete = []
    for col in cols:
        if "deleted" in col:
            to_delete.append(col)
        else:
            not_delete.append(col)
    return to_delete, not_delete


def get_tablename(query):

    cols = list((query & False).fetch(format="frame").reset_index().columns)
    for col in cols:
        if "name" in col and col != "full_name":
            return col


# some datajoint functions seem to be broken and require user input
class SimulateEnterContext:
    def __enter__(self):
        self.original_stdin = sys.stdin
        self.fake_input = StringIO("\n")  # '\n' simulates hitting Enter
        sys.stdin = self.fake_input
        self.original_stdout = sys.stdout
        sys.stdout = StringIO()

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdin = self.original_stdin
        self.fake_input.close()
        sys.stdout = self.original_stdout



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.external_utils import schedule_spikesorting, spikesorting_progress
import numpy as np


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Electrophysiology")
        st.subheader("Schedule Spikesorting")

        st.divider()

        st.subheader("Select the data you want to sort")

        # get user to interactively select session
        _, session_dict = dropdown_query_table(
            tables=tables,
            subtables={
                "Recording": tables["Recording"]
                & (
                    tables["Recording"]
                    * (tables["SortingParams"] & tables["ProbeInsertion"])
                    - tables["SpikeSorting"]
                ).proj()
            },
            username=username,
            headless=True,
        )

        if session_dict == None:
            st.error(
                """You can't perform any spikesorting yet as you haven't inserted any recordings into the database, or all current recordings have already been spikesorted."""
            )
            st.warning("Please go to the insert tab and add some data.")

        else:

            st.divider()

            # check parameters section
            st.subheader("Check your parameters")

            # query database to show parameter sets for these experiments
            query = (
                (
                    tables["Recording"]
                    & (
                        tables["Recording"]
                        * (
                            tables["SortingParams"]
                            & {"manually_sorted": "False"}
                            & tables["ProbeInsertion"]
                        )
                        - tables["SpikeSorting"]
                        & session_dict
                    ).proj()
                )
                .aggr(
                    (
                        tables["SortingParams"] * tables["Recording"]
                        - tables["SpikeSorting"]
                    ).proj(),
                    number_params="count(*)",
                )
                .proj("number_params")
            )
            st.text("")

            if st.button("Check"):

                # can only spikesort your own data
                if session_dict["experimenter"] != username:

                    st.text("")

                    st.error("You can only perform computations on your own data")

                # need at least one parameter set
                elif len(query) == 0:

                    st.text("")

                    st.warning("There is nothing to be spikesorted.")
                    st.info(
                        """This could be because you haven't yet inserted spikesorting parameters or probeinsertions for these animals, or all available recordings and parameter sets have already been computed."""
                    )

                else:

                    df = query.fetch()

                    st.write("Sessions to be spikesorted:")

                    st.dataframe(df)

                    numjobs = df["number_params"].sum()

                    st.write(f"Total number of spikesorting jobs to run: {numjobs}")

            st.divider()

            # schedule computation section
            st.subheader("Schedule computation")

            # ask for user password for scheduling the job
            password = st.text_input(
                "Please enter your cluster password", type="password"
            )

            st.text("")

            # Spikesort
            if st.button("Spikesort"):

                # can only spikesort your own data
                if session_dict["experimenter"] != username:

                    st.error("You can only perform computations on your own data")

                # need at least one parameter set
                elif len(query) == 0:

                    st.error("There is no data to spikesort.")

                else:

                    numjobs = query.fetch()["number_params"].sum()

                    # correction since datajoint returns np.int64 which isn't serialisable
                    session_dict = {
                        key: (int(val) if isinstance(val, np.integer) else val)
                        for key, val in session_dict.items()
                    }

                    try:
                        # send job
                        schedule_spikesorting(session_dict, username, password, numjobs)

                        # success message
                        st.success("Spikesorting job sent to cluster!")
                        st.info("You will receive an email once your job is completed")

                    except Exception as e:

                        print(e)
                        st.error("Error submitting job to cluster")

            st.text("")

        # if there are any downlaods this session
        if "spikesort_jobs" in st.session_state:

            # button which shows spikesort statuses
            if st.button("Check spikesorting progress"):

                spikesorting_progress()



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.visualisation_utils import (
    plot_kinematics,
    choose_kinematics,
    session_length,
)


def show(username, tables):

    st.title("Visualisation")
    st.subheader("Interactively visualise the kinematics of a session")
    st.divider()

    # check if there is any data in the database
    data = len(tables["Kinematics"] & True) > 0

    if not data:
        st.error("""There's no data in the database to visualise""")
        st.warning(
            "Please make sure you have either some spikesorted data or some behavioural data to visualise"
        )

    else:

        col1, col2 = st.columns([0.4, 0.6], gap="large")

        with col1:

            st.markdown("**Select session**")

            # only select session
            session_tables = {"Session": tables["Session"]}

            # get user to interactively insert table attributes
            tablename, session_dict = dropdown_query_table(
                tables, session_tables, username, headless=True
            )

            if "session_id" in session_dict.keys():

                # get user to select the behavioural objects they want to visualise
                selected_object = choose_kinematics(tables, session_dict)

                # pull spike trains
                if session_dict and selected_object:

                    st.divider()
                    st.markdown("**Time range**")

                    # get the user to select the time window they want to visualise
                    max_time = session_length(session_dict)
                    max_time = int(max_time + 0.5)
                    timerange = st.slider(
                        "Select the time range you want to visualise",
                        0,
                        max_time,
                        (0, max_time),
                    )

                    st.divider()
                    if st.button("Plot"):

                        with col2:

                            # create the interactive plotly figure (uses webgl gpu rendering!)
                            fig = plot_kinematics(selected_object, timerange)

                            # customize the display mode bar
                            config = {
                                "displayModeBar": True,
                                "displaylogo": False,
                                "modeBarButtonsToRemove": [
                                    "sendDataToCloud",
                                    "autoScale2d",
                                    "select2d",
                                    "lasso2d",
                                    "toImage",
                                ],
                            }

                            # plot the figure
                            st.plotly_chart(
                                fig, use_container_width=True, config=config
                            )

                else:
                    st.warning("No data to visualise")



import streamlit as st
from antelope.utils.streamlit_utils import (
    server_directory_browser,
    dropdown_insert_table,
)
from antelope.utils.antelope_utils import insert_nwb, check_animals, check_nwb
from antelope.utils.multithreading_utils import behaviour_thread_pool
import datajoint as dj
import pandas as pd


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Import Behaviour Data")

        st.divider()

        st.subheader(
            "Select the session and rig you want to upload behavioural data for"
        )

        tablename, world = dropdown_insert_table(
            tables, {"World": tables["World"]}, username, headless=True
        )

        # if no features in database raise warning
        if tablename == None:
            st.warning(
                """You can't insert any behavioural data because you either don't have the necessary upstream entries or all your sessions already have behavioural data inserted."""
            )

        else:

            st.text("")
            st.info(
                "Select `dlc_training = True` if you want to use this session to train your deeplabcut model."
            )

            # pull behaviour rig
            rig_dict = (tables["BehaviourRig"] & world).fetch1("rig_json")

            # compute the animals in the rig
            animals = check_animals(rig_dict)

            # loop through the animals, getting the user to select which animal it is
            status = True
            animal_keys = {}
            for animal in animals:

                st.divider()
                st.subheader(f"Select animal {animal}")

                # query what animals are in the database
                query = (
                    tables["Animal"]
                    & world
                    & dj.AndList([dj.Not(i) for i in animal_keys.values()])
                )
                animal_select_dict = {
                    i["animal_name"]: i
                    for i in query.proj("animal_name").fetch(as_dict=True)
                }

                # if no animals left, raise warning
                if len(animal_select_dict) == 0:
                    st.warning("No animals left to select")
                    status = False
                    break
                else:

                    # get user to select animal
                    animal_key = animal_select_dict[
                        st.selectbox("Select animal", list(animal_select_dict.keys()))
                    ]
                    del animal_key["animal_name"]
                    animal_keys[animal] = animal_key

            if status:

                # get user to select NWB file
                st.divider()

                st.subheader("Select the NWB file you want to upload")

                nwbpath = server_directory_browser("Select nwb file", "nwb")

                st.divider()

                if st.button("Insert"):

                    # check user only inserting their own data
                    if world["experimenter"] == username:

                        # check nwb file is valid
                        if check_nwb(world, tables, nwbpath):

                            # in separate thread if it has a video, otherwise main thread
                            rig_json = (tables["BehaviourRig"] & world).fetch1(
                                "rig_json"
                            )
                            if ("videos" in rig_json.keys()) and (
                                len(rig_json["videos"]) > 0
                            ):

                                # retrieve thread pool
                                behave_thread_pool = behaviour_thread_pool()

                                # submit job to thread pool
                                query_name = "-".join(
                                    [
                                        key + "-" + str(world[key])
                                        for key in ["experiment_id", "session_id"]
                                    ]
                                )
                                future = behave_thread_pool.submit(
                                    insert_nwb, *(world, animal_keys, nwbpath)
                                )
                                st.session_state.behaviour_futures.append(
                                    (future, query_name)
                                )

                                st.text("")
                                st.success("Data insert in progress")

                            else:

                                insert_nwb(world, animal_keys, nwbpath)
                                st.text("")
                                st.success("Data insert success")

                        else:
                            st.text("")
                            st.error("Invalid NWB file")

                    # otherwise print error
                    else:

                        st.text("")
                        st.error("You can only insert your own data!")

            # add a button which shows behaviour statuses
            # uses all behaviours in current session stored in session state
            if ("behaviour_futures" in st.session_state) and (
                len(st.session_state.behaviour_futures) > 0
            ):
                st.text("")

                if st.button("Check upload progress"):

                    st.write("upload statuses:")

                    # initialise data
                    display_futures = []

                    # display job progress
                    for future, query_name in st.session_state.behaviour_futures:

                        # compute statuses
                        if future.done():
                            if future.exception():
                                print(future.exception())
                                status = "upload error"
                            else:
                                status = "upload success"
                        else:
                            status = "upload in progress"

                        display_futures.append((query_name, status))

                    # make dataframe to display
                    df = pd.DataFrame(display_futures, columns=["Query", "Status"])

                    # show dataframe
                    st.dataframe(df, hide_index=True)



import inspect
import pandas as pd
from pathlib import Path
from antelope.utils.hash_utils import hash_tables, list_to_datajoint, code_hash
import json


def antelope_analysis(cls):
    """
    This is a class decorator used to reduce boilerplate code in analysis functions.
    """

    # initialise with connection
    def __init__(self, conn):
        super(cls, self).__init__(conn)

    cls.__init__ = __init__

    # inherit from AnalysisBase
    class Child(cls, AnalysisBase):
        __doc__ = cls.__doc__
        pass

    return Child


class AnalysisBase:
    name: str
    query: str
    returns: dict
    args: dict
    hidden: bool = False

    def __init__(self, conn):
        self.conn = conn
        self.tables = conn.tables

    def run(self, *args):
        raise NotImplementedError

    def get_directory(self):
        for i in self.__class__.__bases__:
            if i.__name__ != "AnalysisBase":
                return i.__module__.split(".")[-1]

    def __call__(self, restriction={}, *args, **kwargs):

        # first, check everything initialized
        if not all([self.name, self.query, self.returns]):
            raise TypeError("Function not properly defined")

        # append built-in restriction
        if hasattr(self, "key"):
            restriction = {**restriction, **self.key}

        # get primary keys from restriction
        if isinstance(self.query, str):
            table = self.tables[self.query].proj()
        elif isinstance(self.query, list):
            table = self.tables[self.query[0]]
            for q in self.query[1:]:
                table = table * self.tables[q].proj()
        dataset = (table & restriction).proj().fetch(as_dict=True)

        # loop through primary keys and run function
        results = []
        for primary_key in dataset:

            result = self.run(primary_key, *args, **kwargs)
            answer = primary_key.copy()
            for i, key in enumerate(self.returns.keys()):
                if len(self.returns.keys()) == 1:
                    answer[key] = result
                else:
                    answer[key] = result[i]
            results.append(answer)

        if len(results) == 1:
            return results[0]
        else:
            return results

    def save_result(self, filepath="./result.pkl", restriction={}, *args, **kwargs):
        """
        Runs the function, and saves the pickled result to disk along with the reproducibility json.
        """

        # get argument dictionary
        full_args = get_full_args_dict(self, *args, **kwargs)

        tables = list_to_datajoint(self.tables, self.hash)

        # hash tables
        data_hash = hash_tables(tables, restriction)
        codehash = code_hash(self)

        # save reproducibility
        reproducibility = {
            "name": self.get_directory() + "." + self.name,
            "restriction": restriction,
            "arguments": full_args,
            "data_hash": data_hash,
            "code_hash": codehash,
        }

        # run function
        result = self(restriction, *args, **kwargs)
        result = pd.DataFrame(result)

        # save result
        result.to_pickle(filepath)

        with open(Path(filepath).with_suffix(".json"), "w") as f:
            json.dump(reproducibility, f, indent=4)

    def reproduce(self, json_path, result_path):
        """
        Loads a reproducibility json and runs the function with the same arguments,
        checking the data hash and code hash.
        """

        with open(json_path, "r") as f:
            reproducibility = json.load(f)

        tables = list_to_datajoint(self.tables, self.hash)

        # hash tables
        data_hash = hash_tables(tables, reproducibility["restriction"])
        codehash = code_hash(self)

        # assert everything is the same
        assert (
            reproducibility["name"] == self.get_directory() + "." + self.name
        ), "Function incorrect."
        assert (
            reproducibility["data_hash"] == data_hash
        ), "Data has changed since last function run."
        assert (
            reproducibility["code_hash"] == codehash
        ), "Code has changed since last function run."
        print("Reproducibility checks passed.")
        print("Running function...")

        # run function
        result = self(reproducibility["restriction"], **reproducibility["arguments"])

        # save result
        result = pd.DataFrame(result)
        result.to_pickle(result_path)

    def check_hash(self, json_path):
        """
        Loads a reproducibility json and runs the function with the same arguments,
        checking the data hash and code hash.
        """

        with open(json_path, "r") as f:
            reproducibility = json.load(f)

        # tables to hash
        tables = list_to_datajoint(self.tables, self.hash)

        # hash tables
        data_hash = hash_tables(tables, reproducibility["restriction"])
        codehash = code_hash(self)

        # check conditions
        if not reproducibility["name"] == self.get_directory() + "." + self.name:
            return "Function incorrect."
        elif not reproducibility["data_hash"] == data_hash:
            return "Data has changed since last function run."
        elif not reproducibility["code_hash"] == codehash:
            return "Code has changed since last function run."
        else:
            return "Reproducibility checks passed."


def get_full_args_dict(func, *args, **kwargs):

    # Get the signature of the function
    sig = inspect.signature(func.run)

    # Create a dictionary of the default values
    defaults = {
        k: v.default
        for k, v in sig.parameters.items()
        if v.default is not inspect.Parameter.empty and k != "restriction"
    }

    # Get the names of all parameters
    param_names = [k for k in sig.parameters.keys() if k != "key"]

    # Update the defaults dictionary with provided args
    args_dict = dict(zip(param_names, args))

    # Combine the defaults, provided args, and provided kwargs
    full_args = {**defaults, **args_dict, **kwargs}

    return full_args



import streamlit as st
import pandas as pd
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.datajoint_utils import (
    form_query,
    query_without_external,
    query_to_str,
    release_computation,
)
from antelope.utils.multithreading_utils import release_thread_pool


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Database Administration")
        st.subheader("Release entry stuck in computation")

        st.divider()

        # just check spikesorting table
        subtables = {key: val for key, val in tables.items() if key == "SpikeSorting"}

        # get input restriction
        tablename, restriction = dropdown_query_table(
            tables, subtables, username=None, in_compute="True", headless=True
        )

        if not tablename:

            st.warning("There are no entries currently in computation.")

        else:

            # form datajoint querystring
            querystring = f"{tablename} & {str(restriction)}"

            # form query
            query = form_query(querystring, tables)

            st.text("")

            if st.button("Search"):

                st.text("")

                if querystring:

                    st.markdown("##### Data in computation:")

                    # query results to display
                    df, number = query_without_external(query)

                    # show table
                    if number == 0:
                        st.warning("No entries to show!")
                    else:
                        st.dataframe(df, hide_index=True)
                        st.text("Number of items: {}".format(number))
                else:
                    st.error("You must enter a query first!")

            st.text("")

            # warning message
            st.info(
                """
            Note that the behaviour of this function depends on whether your data is stuck in computation during spikesorting, or while uploading the results of your manual curation.\n
            In the former case, this amounts to a true delete, as there is no data that needs to be saved, while in the latter case, it simply sets the 'in_compute' attribute to False, so that the job can be rerun."""
            )
            st.text("")
            st.warning(
                """
            Before releasing an entry from computation, please check the logs to see why the computation failed, and make the appropriate modifications, to avoid wasting computational resources.\n
            Please also try to ensure you don't release data that is actually currently in computation, as opposed to having failed computation, as this will cause your job to fail, wasting resources.\n
            Common causes of jobs failing are corrupted data, misentered parameters, or the job timing out."""
            )

            st.text("")

            status = (
                st.text_input(
                    label="Please enter your password to confirm computation release:",
                    type="password",
                )
                == st.session_state.password
            )

            st.text("")

            # release from computation
            if st.button("Release computation"):

                st.text("")

                if querystring:

                    st.text("")

                    # delete only if password correct
                    if status:

                        # check how many entries to release
                        num_release = len(query)

                        # if number releases over 1000, run in separate thread
                        if num_release > 1000:

                            # retrieve thread pool
                            rel_thread_pool = release_thread_pool()

                            # make name for query
                            query_name = query_to_str(form_query(querystring, tables))

                            # submit job to thread pool
                            future = rel_thread_pool.submit(
                                release_computation, tablename, restriction
                            )

                            # append future to session state
                            st.session_state.release_futures.append(
                                (future, query_name)
                            )

                            st.success("Release computation in progress!")

                        # otherwise run in main thread
                        else:

                            # release data
                            release_computation(tablename, restriction)

                            st.success("Data released from computation!")

                    # incorrect password
                    else:
                        st.error("Password incorrect.")

                else:
                    st.error("You must enter a query first!")

            st.text("")

            # add a button which shows download statuses
            # uses all downlaods in current session stored in session state
            # if there are any downloads this session
            if "release_futures" in st.session_state:

                if st.button("Check release computation progress"):

                    st.write("Release computation statuses:")

                    # initialise data
                    display_futures = []

                    # display job progress
                    for future, query_name in st.session_state.release_futures:

                        # compute statuses
                        if future.done():
                            if future.exception():
                                print(future.exception())
                                status = "release computation error"
                            else:
                                status = "release computation success"
                        else:
                            status = "release computation in progress"

                        display_futures.append((query_name, status))

                    # make dataframe to display
                    df = pd.DataFrame(display_futures, columns=["Query", "Status"])

                    # show dataframe
                    st.dataframe(df, hide_index=True)

                # if there are no downloads in this session
                else:
                    st.write("No downloads underway.")



import socket
from random import randint


def is_port_available(port: int, host: str) -> bool:
    """Check if a port is available on the given host."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        return sock.connect_ex((host, port)) != 0


def find_available_port(
    min_port: int = 10000,
    max_port: int = 65535,
    max_tries: int = 50,
    host: str = "localhost",
) -> int:
    """Find an available port on the given host."""
    for _ in range(max_tries):
        selected_port = randint(min_port, max_port)
        if is_port_available(selected_port, host):
            return selected_port
    raise RuntimeError("Unable to find an available port.")


if __name__ == "__main__":
    print(find_available_port())



from antelope.connection.connect import dbconnect
from antelope.connection import import_schemas
from antelope.utils.os_utils import get_config
from antelope.utils.analysis_utils import import_analysis, reload_analysis
from IPython import embed
import getpass
from types import SimpleNamespace
import antelope.scripts.hold_conn as hold


def run():

    # first check config file exists
    config = get_config()
    if config is None:
        print("Config file not found.")
        print("Please run `antelope-config` to generate a configuration file.")
        exit()

    # connect to database
    username = input("Please enter your username: ")
    password = getpass.getpass("Please enter your password: ")

    global conn
    conn = dbconnect(username, password)
    global tables
    tables = import_schemas.schema(conn)

    if hold.conn is None:
        hold.conn = conn
    if hold.tables is None:
        hold.tables = tables

    for key, val in tables.items():
        globals()[key] = val

    # import analysis functions
    analysis_functions = import_analysis(conn, tables)
    for folder, fcts in analysis_functions.items():
        # make class
        if folder not in globals():
            globals()[folder] = SimpleNamespace()
        for fct in fcts:
            setattr(globals()[folder], fct.name, fct)

    def reload():
        """
        Global function to relaod all analysis functions interactively
        """
        analysis_functions = reload_analysis(conn, tables)
        for folder, fcts in analysis_functions.items():
            # make class
            if folder not in globals():
                globals()[folder] = SimpleNamespace()
            for fct in fcts:
                setattr(globals()[folder], fct.name, fct)

    # start ipython session
    embed()



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_insert_table
from datajoint.errors import DuplicateError


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Electrophysiology")
        st.subheader("Make Probe Insertion")

        st.divider()

        # get user to interactively insert table attributes
        tablename, insert_dict = dropdown_insert_table(
            tables,
            {"ProbeInsertion": tables["ProbeInsertion"]},
            username,
            headless=True,
        )

        # if upstream tables not populated, print error
        if tablename is None:

            st.text("")
            st.error(
                """You can't make a probe insertion yet as you haven't got entries in the necessary upstream tables."""
            )
            st.warning(
                "Please make sure you have inserted an animal and a probe before attempting to insert a probe insertion."
            )
            st.stop()

        st.text("")

        if st.button("Insert"):

            try:

                # check user only inserting their own data
                if insert_dict["experimenter"] == username:

                    tables[tablename].insert1(insert_dict)

                    # show success message
                    st.text("")
                    st.success("Data uploaded to database!")

                # otherwise print error
                else:

                    st.text("")
                    st.error("You can only insert your own data!")

            except DuplicateError:

                st.error("This animal already has a probe insertion!")



import jsonschema
from pathlib import Path
from antelope.connection import connect, import_schemas
from antelope.utils.analysis_utils import import_analysis, instantiate_function
from pynwb import NWBHDF5IO, NWBFile, TimeSeries
from pynwb.file import Subject
from pynwb.image import ImageSeries
from pynwb.misc import IntervalSeries
from pynwb.behavior import (
    BehavioralTimeSeries,
    BehavioralEpochs,
    Position,
    SpatialSeries,
)
import numpy as np
import tempfile
import yaml
import json
import shutil
import zipfile
import sys
import streamlit as st
import pandas as pd
from functools import partial
import inspect
import traceback
import os
from uuid import uuid4

"""
It's important to mention how our naming conventions work for animals.
Some behavioural events could belong to a specific animal a priori.
Other events, we won't know which animal they belong to until we perform tracking with deeplabcut.
In the first case, the animal will be numbered in the rig specification. For example, a setup with 2 animals will have
features with source_type as a dictionary, with the key 'acquisition' and values 1 or 2.
In the second case, the value of the key 'acquisition' will be -1.
Such objects in the database will initally refer to neither world or self.
After tracking, we will be able to split the arrays, splitting the object table, and referencing the correct self entries.
"""


def validate_behaviour_rig(data):
    """
    Function validates an input behavior rig specification against our custom schema
    :param data: dict
    :return: Bool
    """

    # first, define the schema
    schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "specification": {"type": "string", "const": "antelope-behaviour"},
            "version": {"type": "string"},
            "reference_point": {"type": "string"},
            "features": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "source": {
                            "type": "object",
                            "properties": {
                                "source_type": {
                                    "type": "string",
                                    "enum": [
                                        "acquisition",
                                        "stimulus",
                                        "processing",
                                        "deeplabcut",
                                    ],
                                },
                                "module": {"type": "string"},
                                "video": {"type": "string"},
                            },
                            "required": ["source_type"],
                        },
                        "ownership": {
                            "type": "object",
                            "properties": {
                                "ownership": {
                                    "type": "string",
                                    "enum": ["world", "self"],
                                },
                                "animal": {"type": "integer"},
                            },
                            "required": ["ownership"],
                        },
                        "data_type": {
                            "type": "string",
                            "enum": ["digital", "analog", "interval", "kinematics"],
                        },
                        "coordinates": {
                            "type": "array",
                            "minItems": 3,
                            "maxItems": 3,
                            "items": {"type": "number"},
                        },
                        "description": {"type": "string"},
                    },
                    "required": ["name", "source", "data_type", "description"],
                },
            },
            "videos": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "description": {"type": "string"},
                        "format": {"type": "string", "enum": ["avi", "mp4", "mov"]},
                        "reference_point": {"type": "string"},
                    },
                    "required": ["name", "description", "format"],
                },
            },
        },
        "required": ["specification", "version", "reference_point", "features"],
    }

    # now validate the input data
    try:
        jsonschema.validate(data, schema)
    except jsonschema.exceptions.ValidationError:
        return "Json file not valid!"

    # also check the names are all unique
    names = [i["name"] for i in data["features"]]
    if len(names) != len(set(names)):
        return "Feature names not unique!"
    else:
        return True


def validate_mask(data):
    """
    Function validates an input behavior rig specification against our custom schema
    :param data: dict
    :return: Bool
    """

    # first, define the schema
    schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "function": {"type": "string"},
            },
            "required": ["name", "description", "function"],
        },
    }

    # now validate the input data
    try:
        jsonschema.validate(data, schema)
    except jsonschema.exceptions.ValidationError:
        return "Masks not valid!"
    else:
        return True


def insert_features(primary_key, rig_json):
    """
    Function populates the features table from the json file,
    leaving the feature empty
    """

    # create list to store insert dicts
    feature_list = []

    for i, feature in enumerate(rig_json["features"]):

        # create insert dict
        insert_feature = {}

        # make the feature fit the datajoint table
        insert_feature["experimenter"] = primary_key["experimenter"]
        insert_feature["behaviourrig_id"] = primary_key["behaviourrig_id"]
        insert_feature["feature_id"] = i + 1
        insert_feature["feature_name"] = feature["name"]
        insert_feature["source_type"] = feature["source"]["source_type"]
        insert_feature["data_type"] = feature["data_type"]
        insert_feature["feature_description"] = feature["description"]
        insert_feature["feature_data"] = None
        insert_feature["feature_deleted"] = "False"

        # append to list
        feature_list.append(insert_feature)

    return feature_list


def upload_rig_json(experimenter, rig_name, rigfile, masklist):
    """
    Function validates the rig json, converts and uploads it along with corresponding features

    Inputs:
        experimenter (str): experimenter username
        rig_name (str): name of the rig
        rig_json (path): path to the rig json file
        masklist (list of dict): list of mask dictionaries, with keys 'name', 'description', 'function'
    """
    import antelope.load_connection as db

    with open(rigfile, "r") as f:
        read_rig = json.load(f)

    # validate the rig file
    assert validate_behaviour_rig(read_rig)
    assert validate_mask(masklist)

    insert_dict = {
        "experimenter": experimenter,
        "behaviourrig_name": rig_name,
        "rig_json": read_rig,
        "masks": masklist,
    }

    with db.BehaviourRig.connection.transaction:

        db.BehaviourRig.insert1(insert_dict)
        primary = {
            key: val
            for key, val in insert_dict.items()
            if key not in ["rig_json", "masks"]
        }
        primary_key = (db.BehaviourRig & primary).proj().fetch1()

        feature_list = insert_features(primary_key, read_rig)

        db.Feature.insert(feature_list)


def make_timestamps(length, sample_rate):
    """
    Function makes timestamps for analog events
    """
    return np.linspace(length) / sample_rate


def check_nwb(world, tables, nwbpath):
    """
    Function reads an NWB file and the database features and checks the nwb file is valid
    """

    # pull the features from the database
    rig_json = (tables["BehaviourRig"] & world).fetch1("rig_json")
    features = rig_json["features"]
    videos = rig_json["videos"]

    try:

        # open nwb file
        with NWBHDF5IO(nwbpath, "r") as io:
            read_nwbfile = io.read()

            video_names = {}

            # check the videos
            for i, video in enumerate(videos):

                # get the video path
                video_name = video["name"]
                assert (
                    video_name in read_nwbfile.acquisition.keys()
                ), f"Video {video} not in NWB file"
                video_relative_path = read_nwbfile.acquisition[
                    video_name
                ].external_file[0]
                video_full_path = nwbpath.parent / Path(video_relative_path)
                assert video_full_path.exists(), f"Video {video_name} not in directory"

            # insert the features and objects
            for feature in features:

                # get the data and timestamps
                if feature["source"]["source_type"] == "acquisition":
                    assert hasattr(read_nwbfile.acquisition[feature["name"]], "data")

                elif feature["source"]["source_type"] == "stimulus":
                    assert hasattr(read_nwbfile.stimulus[feature["name"]], "data")

                elif feature["source"]["source_type"] == "processing":
                    assert hasattr(
                        read_nwbfile.processing[feature["source"]["module"]][
                            feature["name"]
                        ],
                        "data",
                    )

    except AssertionError:
        return False

    else:
        return True


def check_streamlit():
    """
    Function to check whether python code is run within streamlit

    Returns
    -------
    use_streamlit : boolean
        True if code is run within streamlit, else False
    """
    try:
        from streamlit.runtime.scriptrunner import get_script_run_ctx

        if not get_script_run_ctx():
            use_streamlit = False
        else:
            use_streamlit = True
    except ModuleNotFoundError:
        use_streamlit = False
    return use_streamlit


def insert_nwb(world, selfs, nwbpath):
    """
    Function reads an NWB file and the database features and inserts the data into the database

    Args:
        session (dict): session primary key
        animals (list of dict): animal primary keys
        nwbpath (Path): path to the nwb file

    Returns:
        None
    """

    # function runs in different process so needs to establish its own connection
    if check_streamlit():
        conn_insert = connect.connect()
        tables_insert = import_schemas.schema(conn_insert)
    elif hasattr(sys, "ps1") or sys.flags.interactive:
        import antelope.scripts.hold_conn as hold

        conn_insert = hold.conn
        tables_insert = hold.tables
    else:
        import antelope.load_connection as db

        conn_insert = getattr(db, "conn")
        tables_insert = {}
        for table in dir(db):
            if not table.startswith("_"):
                tables_insert[table] = getattr(db, table)

    # pull the features from the database
    rig_json = (tables_insert["BehaviourRig"] & world).fetch1("rig_json")
    masks = (
        (tables_insert["MaskFunction"] & world)
        .proj("maskfunction_name")
        .fetch(as_dict=True)
    )
    features = rig_json["features"]
    videos = rig_json["videos"]

    with conn_insert.transaction:

        # insert world
        tables_insert["World"].insert1(world)
        if "world_deleted" in world.keys():
            del world["world_deleted"]
        if "dlc_training" in world.keys():
            del world["dlc_training"]
        behaviourrig_id = world["behaviourrig_id"]
        selfkey = world.copy()
        del selfkey["behaviourrig_id"]

        # insert selfs
        for self in selfs.values():
            tables_insert["Self"].insert1(
                {**self, "session_id": world["session_id"], "self_deleted": "False"}
            )

        # open nwb file
        with NWBHDF5IO(nwbpath, "r") as io:
            read_nwbfile = io.read()

            video_names = {}

            # insert the videos
            for i, video in enumerate(videos):

                # get the video path
                video_name = video["name"]
                video_relative_path = read_nwbfile.acquisition[
                    video_name
                ].external_file[0]
                if hasattr(read_nwbfile.acquisition[video_name], "timestamps"):
                    timestamps = read_nwbfile.acquisition[video_name].timestamps[:]
                else:
                    timestamps = None
                video_full_path = nwbpath.parent / Path(video_relative_path)
                video_names[video_name] = i + 1

                # insert the video
                tables_insert["Video"].insert1(
                    {
                        **world,
                        "video_id": i + 1,
                        "video": str(video_full_path),
                        "timestamps": timestamps,
                        "video_deleted": "False",
                    },
                    allow_direct_insert=True,
                )

            # insert the features and objects
            for feature in features:

                # calculate feature id from feature name
                query = (
                    tables_insert["Feature"] & world & {"feature_name": feature["name"]}
                )
                feature_id, feature_name = query.fetch1("feature_id", "feature_name")

                # compute if world or self
                if feature["ownership"]["ownership"] == "world":
                    object_type = "World"
                    animal_id = None

                elif feature["ownership"]["ownership"] == "self":
                    object_type = "Self"
                    animal_id = selfs[feature["ownership"]["animal"]]["animal_id"]

                # feature dict to insert
                world["behaviourrig_id"] = behaviourrig_id
                insert_dict = {
                    "feature_id": feature_id,
                    **world,
                    "object_name": feature_name,
                    "object_type": object_type,
                    "animal_id": animal_id,
                    "object_deleted": "False",
                }

                # insert feature dict
                tables_insert["Object"].insert1(insert_dict, allow_direct_insert=True)

                # does the data need units
                if feature["data_type"] in ["digital", "analog"]:
                    need_unit = True
                else:
                    need_unit = False

                # does the data need coordinates
                if feature["data_type"] in ["digital", "analog", "interval"]:
                    coordinates = {
                        "x_coordinate": feature["coordinates"][0],
                        "y_coordinate": feature["coordinates"][1],
                        "z_coordinate": feature["coordinates"][2],
                    }

                # get the data and timestamps
                if feature["source"]["source_type"] == "acquisition":
                    data = read_nwbfile.acquisition[feature["name"]].data[:]
                    if hasattr(read_nwbfile.acquisition[feature["name"]], "timestamps"):
                        timestamps = read_nwbfile.acquisition[
                            feature["name"]
                        ].timestamps[:]
                    else:
                        sample_rate = read_nwbfile.acquisition[feature["name"]].rate
                        timestamps = make_timestamps(len(data), sample_rate)
                    if need_unit:
                        unit = read_nwbfile.acquisition[feature["name"]].unit

                elif feature["source"]["source_type"] == "stimulus":
                    data = read_nwbfile.stimulus[feature["name"]].data[:]
                    if hasattr(read_nwbfile.stimulus[feature["name"]], "timestamps"):
                        timestamps = read_nwbfile.stimulus[feature["name"]].timestamps[
                            :
                        ]
                    else:
                        sample_rate = read_nwbfile.stimulus[feature["name"]].rate
                        timestamps = make_timestamps(len(data), sample_rate)
                    if need_unit:
                        unit = read_nwbfile.stimulus[feature["name"]].unit

                elif feature["source"]["source_type"] == "processing":
                    data = read_nwbfile.processing[feature["source"]["module"]][
                        feature["name"]
                    ].data[
                        :
                    ]  # need to check if this is how we access data in modules
                    if hasattr(
                        read_nwbfile.processing[feature["source"]["module"]][
                            feature["name"]
                        ],
                        "timestamps",
                    ):
                        timestamps = read_nwbfile.processing[
                            feature["source"]["module"]
                        ][feature["name"]].timestamps[:]
                    else:
                        sample_rate = read_nwbfile.processing[
                            feature["source"]["module"]
                        ][feature["name"]].rate
                        timestamps = make_timestamps(len(data), sample_rate)
                    if need_unit:
                        unit = read_nwbfile.processing[feature["source"]["module"]][
                            feature["name"]
                        ].unit

                # compute if the event is digital or analog
                if feature["data_type"] == "digital":
                    data_insert = {
                        "feature_id": feature_id,
                        **world,
                        "animal_id": animal_id,
                        "data": data,
                        "timestamps": timestamps,
                        **coordinates,
                        "unit": unit,
                        "digitalevents_name": feature_name,
                        "digitalevents_deleted": "False",
                    }
                    tables_insert["DigitalEvents"].insert1(
                        data_insert, allow_direct_insert=True
                    )

                elif feature["data_type"] == "analog":
                    data_insert = {
                        "feature_id": feature_id,
                        **world,
                        "animal_id": animal_id,
                        "data": data,
                        "timestamps": timestamps,
                        **coordinates,
                        "unit": unit,
                        "analogevents_name": feature_name,
                        "analogevents_deleted": "False",
                    }
                    tables_insert["AnalogEvents"].insert1(
                        data_insert, allow_direct_insert=True
                    )

                elif feature["data_type"] == "interval":
                    data_insert = {
                        "feature_id": feature_id,
                        **world,
                        "animal_id": animal_id,
                        "data": data,
                        "timestamps": timestamps,
                        **coordinates,
                        "intervalevents_name": feature_name,
                        "intervalevents_deleted": "False",
                    }
                    tables_insert["IntervalEvents"].insert1(
                        data_insert, allow_direct_insert=True
                    )

                elif feature["data_type"] == "kinematics":
                    if feature["source"]["source_type"] == "processing":

                        # get the video
                        video_name = feature["source"]["video"]
                        video_id = video_names[video_name]

                        # insert the data
                        data_insert = {
                            "feature_id": feature_id,
                            **world,
                            "video_id": video_id,
                            "animal_id": animal_id,
                            "data": data,
                            "timestamps": timestamps,
                            "kinematics_name": feature_name,
                            "kinematics_deleted": "False",
                        }
                        tables_insert["Kinematics"].insert1(
                            data_insert, allow_direct_insert=True
                        )

            # compute the masks
            for i, mask in enumerate(masks):

                mask_function = instantiate_function(mask, tables_insert)
                name = mask["maskfunction_name"]
                del mask["maskfunction_name"]

                # run the function
                output = mask_function(world)
                data, timestamps = output["data"], output["timestamps"]

                # insert the mask
                tables_insert["Mask"].insert1(
                    {
                        **world,
                        "mask_id": mask["maskfunction_id"],
                        "data": data,
                        "timestamps": timestamps,
                        "mask_name": mask["maskfunction_name"],
                        "mask_deleted": "False",
                    },
                    allow_direct_insert=True,
                )


def recompute_masks(restriction):

    # function runs in different process so needs to establish its own connection
    if check_streamlit():
        conn_insert = connect.connect()
        tables_insert = import_schemas.schema(conn_insert)
    elif hasattr(sys, "ps1") or sys.flags.interactive:
        import antelope.scripts.hold_conn as hold

        conn_insert = hold.conn
        tables_insert = hold.tables
    else:
        import antelope.load_connection as db

        conn_insert = getattr(db, "conn")
        tables_insert = {}
        for table in dir(db):
            if not table.startswith("_"):
                tables_insert[table] = getattr(db, table)

    # pull the world from the database
    worlds = (tables_insert["World"] & restriction).proj().fetch(as_dict=True)

    for world in worlds:

        # pull the masks from the database
        masks = (
            (
                tables_insert["MaskFunction"] * tables_insert["World"]
                - tables_insert["Mask"]
                & world
            )
            .proj("maskfunction_name")
            .fetch(as_dict=True)
        )

        # compute the masks
        for mask in masks:

            mask_function = instantiate_function(mask, tables_insert)
            name = mask["maskfunction_name"]
            del mask["maskfunction_name"]

            # run the function
            output = mask_function(world)
            data, timestamps = output["data"], output["timestamps"]

            # insert the mask
            if tables_insert["Mask"] & mask:
                tables_insert["Mask"].update1(
                    {
                        **mask,
                        "data": data,
                        "timestamps": timestamps,
                        "mask_deleted": "False",
                    }
                )
            else:
                tables_insert["Mask"].insert1(
                    {
                        **mask,
                        "data": data,
                        "timestamps": timestamps,
                        "mask_name": name,
                        "mask_deleted": "False",
                    },
                    allow_direct_insert=True,
                )


def check_animals(rig_dict):
    """
    Function reads our custom rig schema and checks how many animals are in the rig
    """
    animals = []
    for feature in rig_dict["features"]:
        if feature["ownership"]["ownership"] == "self":
            animals.append(feature["ownership"]["animal"])

    # return the unique animals in order
    animals = sorted(list(set(animals)))
    animals = [i for i in animals if i != -1]
    return animals


def extract_frames(conf, algo):

    import deeplabcut

    deeplabcut.extract_frames(conf, algo=algo, userfeedback=False)


def check_dlc_folder(key, tables):
    """
    Function checks if a dlc folder exists and is valid.
    """

    name = f"{key['experimenter']}_{key['behaviourrig_id']}_{key['experiment_id']}_{key['dlcmodel_id']}"
    tmpdir = Path(tempfile.gettempdir())
    dlc_dir = tmpdir / "antelope_dlc"
    folder = dlc_dir / name

    # if there's already a dlc folder
    if folder.exists():
        for i in folder.iterdir():
            if i.name == "videos":
                continue
            else:
                conf = i / "config.yaml"
                if conf.exists():
                    with open(conf, "r") as f:
                        config_dict = yaml.safe_load(f)
                    # need to check if all videos are downloaded and match database
                    if all(
                        [Path(vid).exists() for vid in config_dict["video_sets"].keys()]
                    ):  # UPDATE THIS SO WE ALSO CHECK AGAINST DATABASE
                        return True, conf

    return False, None


def create_dlc_folder(key):
    """
    This function creates a dlc folder for a specific experiment.
    If it already exists, it checks the videos, and if they all match, it returns the config path.
    Otherwise, it deletes the DLC folder, downloads the remaining videos, makes the DLC folder and returns the config path.
    """

    # function runs in different process so needs to establish its own connection
    conn_dlc = connect.connect()
    tables_dlc = import_schemas.schema(conn_dlc)

    name = f"{key['experimenter']}_{key['behaviourrig_id']}_{key['experiment_id']}_{key['dlcmodel_id']}"
    tmpdir = Path(tempfile.gettempdir())
    dlc_dir = tmpdir / "antelope_dlc"
    folder = dlc_dir / name
    folder.mkdir(parents=True, exist_ok=True)
    video_path = folder / "videos"
    video_path.mkdir(parents=True, exist_ok=True)

    # if there's already a dlc folder
    for i in folder.iterdir():
        if i.name == "videos":
            continue
        else:
            conf = i / "config.yaml"
            if conf.exists():
                with open(conf, "r") as f:
                    config_dict = yaml.safe_load(f)
                # need to check if all videos are downloaded and match database
                if all(
                    [Path(vid).exists() for vid in config_dict["video_sets"].keys()]
                ):  # UPDATE THIS SO WE ALSO CHECK AGAINST DATABASE
                    return conf
                else:
                    shutil.rmtree(i)
                    break

    # download the videos
    key["dlc_training"] = "True"
    videos = (tables_dlc["Video"] * tables_dlc["World"] & key).fetch(
        download_path=video_path
    )
    video_list = list(videos["video"])

    import deeplabcut

    conf = deeplabcut.create_new_project(
        name,
        key["experimenter"],
        video_list,
        working_directory=folder,
        copy_videos=False,
    )

    # read the config file
    with open(conf, "r") as f:
        config_dict = yaml.safe_load(f)

    # get the behaviour dict
    behaviour = (tables_dlc["BehaviourRig"] & key).fetch1("rig_json")

    bodyparts = []
    for feature in behaviour["features"]:
        if (
            "source_type" in feature["source"]
            and feature["source"]["source_type"] == "deeplabcut"
        ):
            bodyparts.append(feature["name"])

    # add to config dict
    config_dict["bodyparts"] = bodyparts
    del config_dict["skeleton"]
    with open(conf, "w") as f:
        yaml.dump(config_dict, f)

    return conf


def edit_params(key, tables):

    import streamlit as st

    name = f"{key['experimenter']}_{key['behaviourrig_id']}_{key['experiment_id']}_{key['dlcmodel_id']}"
    conf = st.session_state.dlc_folder[name]
    # read the config file
    with open(conf, "r") as f:
        config_dict = yaml.safe_load(f)

    # get user to interactively select options
    st.divider()
    frames = st.slider(
        "Select the range from which you want to extract training frames",
        0.0,
        1.0,
        (0.0, 1.0),
        step=0.01,
    )
    st.text("The frame range is a fraction of the total video length.")
    model = st.selectbox(
        "Select which model to use",
        [
            "resnet_50",
            "resnet_101",
            "resnet_152",
            "mobilenet_v2_1.0",
            "mobilenet_v2_0.75",
            "mobilenet_v2_0.5",
            "mobilenet_v2_0.35",
        ],
    )
    st.text(
        "Resnet_50 should be adequate for most purposes. Please read the DeepLabCut documentation for more information."
    )
    numframes = st.number_input(
        "Enter the number of frames you want to annotate", min_value=10, value=20
    )
    st.text("The more frames you annotate, the better the model will be.")
    training_fraction = st.number_input(
        "Enter the fraction of frames to use for training",
        min_value=0.0,
        max_value=1.0,
        value=0.95,
    )
    st.text(
        "We will use a small portion of your annotated frames to assess the performance of the model."
    )
    p_cutoff = st.number_input(
        "Enter the p-cutoff value", min_value=0.0, max_value=1.0, value=0.6
    )
    st.text(
        "This specifies the threshold of the likelihood and helps distinguish likely body parts from uncertain ones."
    )

    # update the config dict
    config_dict["start"] = frames[0]
    config_dict["stop"] = frames[1]
    config_dict["default_net_type"] = model
    config_dict["numframes2pick"] = numframes
    config_dict["TrainingFraction"] = [training_fraction]
    config_dict["pcutoff"] = p_cutoff

    st.divider()
    augmenter = st.selectbox(
        "Select the augmentation type", ["imgaug", "tensorpack", "scalecrop"]
    )
    st.text(
        "Please read the DeepLabCut documentation for more information on augmentation methods."
    )
    max_snapshots_to_keep = st.number_input(
        "Enter the maximum number of snapshots to keep",
        min_value=1,
        max_value=20,
        value=5,
    )
    st.text("This is the number of snapshots of the model to keep during training.")
    displayiters = st.number_input(
        "Enter the number of iterations between displaying the model",
        min_value=100,
        max_value=10000,
        value=1000,
    )
    st.text(
        "This is the number of iterations between displaying the model performance."
    )
    saveiters = st.number_input(
        "Enter the number of iterations between saving the model",
        min_value=1000,
        max_value=100000,
        value=50000,
    )
    st.text("This is the number of iterations between saving the model.")
    maxiters = st.number_input(
        "Enter the maximum number of iterations",
        min_value=1000,
        max_value=5000000,
        value=1000000,
    )

    train_dict = {
        "augmenter": augmenter,
        "max_snapshots_to_keep": max_snapshots_to_keep,
        "displayiters": displayiters,
        "saveiters": saveiters,
        "maxiters": maxiters,
    }

    params_dict = {"config": config_dict, "compute": train_dict}

    return params_dict


def upload_dlc(key):
    """
    Function zips the labelled frames folder and uploads to the database along with the config json
    """

    # function runs in different process so needs to establish its own connection
    conn_dlc = connect.connect()
    tables_dlc = import_schemas.schema(conn_dlc)

    name = f"{key['experimenter']}_{key['behaviourrig_id']}_{key['experiment_id']}_{key['dlcmodel_id']}"
    conf = st.session_state.dlc_folder[name]
    comp = conf.parent / "compute.json"

    # read the config file
    with open(conf, "r") as f:
        config_dict = yaml.safe_load(f)
    with open(comp, "r") as f:
        compute_dict = json.load(f)
    params_dict = {"config": config_dict, "compute": compute_dict}

    # zip the labelled frames folder
    labelled_frames = conf.parent / "labeled-data"
    target = conf.parent.parent / f"{name}_labelled_frames"
    shutil.make_archive(target, "zip", labelled_frames)
    labeled_zip = target.with_suffix(".zip")

    # upload to database
    key["dlcparams"] = params_dict
    key["labelled_frames"] = str(labeled_zip)
    key["labelledframes_in_compute"] = "False"
    key["labelledframes_deleted"] = "False"
    del key["behaviourrig_deleted"]

    tables_dlc["LabelledFrames"].insert1(key, allow_direct_insert=True)


def select_masks(tables):
    """
    Function displays a streamlit data editor that allows the user to select the masks they want to compute
    """

    # pull available functions
    fcts = import_analysis(tables["Experimenter"].connection, tables)
    fctsls = [
        f"{i}.{j.name}" for i in fcts for j in fcts[i] if hasattr(j, "mask") and j.mask
    ]

    if len(fctsls) == 0:
        st.error("No mask functions available!")
        return

    default = [{"enabled": False, "description": "", "function": fctsls[0]}]
    edited_masks = st.data_editor(
        default,
        use_container_width=True,
        num_rows="dynamic",
        column_config={
            "enabled": st.column_config.CheckboxColumn(help="Enable mask computation"),
            "function": st.column_config.SelectboxColumn(
                help="Select your function to compute your masks",
                required=True,
                options=fctsls,
            ),
            "description": "description",
        },
    )
    enabled_masks = [i for i in edited_masks if i["enabled"]]
    return enabled_masks


def insert_masks(key, masks, id, tables):
    """
    Function makes dataframe out of masks
    """
    df = pd.DataFrame(masks)
    del df["enabled"]
    df["experimenter"] = key["experimenter"]
    df["behaviourrig_id"] = key["behaviourrig_id"]
    df.rename(
        columns={"name": "mask_name", "description": "maskfunction_description"},
        inplace=True,
    )
    df["mask_id"] = range(id, id + len(df))
    df["maskfunction_deleted"] = "False"

    new_serialise = partial(serialise_function, tables=tables)
    df["mask_function"] = df["function"].apply(new_serialise)
    new_name = partial(function_name, tables=tables)
    df["maskfunction_name"] = df["function"].apply(new_name)

    del df["function"]

    return df


def serialise_function(function_name, tables):
    # from a functions folder.name format, return json function code

    fcts = import_analysis(tables["Experimenter"].connection, tables)
    for folder in fcts:
        if function_name.split(".")[0] == folder:
            for fct in fcts[folder]:
                if fct.name == function_name.split(".")[1]:
                    function = fct
                    break
            break
    code = inspect.getsource(function.__class__.__bases__[0])

    return json.dumps(code)


def function_name(function_name, tables):
    # from a functions folder.name format, return class name

    fcts = import_analysis(tables["Experimenter"].connection, tables)
    for folder in fcts:
        if function_name.split(".")[0] == folder:
            for fct in fcts[folder]:
                if fct.name == function_name.split(".")[1]:
                    function = fct
                    break
            break
    name = function.__class__.__bases__[0].__name__
    return name


def display_dlc_images(DLCModel, key):
    """
    Function displays the train and test images for a DLCModel
    """

    # first make temp directory
    name = f"{key['experimenter']}_{key['behaviourrig_id']}_{key['experiment_id']}_{key['dlcmodel_id']}"
    tmpdir = Path(tempfile.gettempdir())
    dlc_dir = tmpdir / "antelope_dlc"
    folder = dlc_dir / name / "evaluation_images"
    folder.mkdir(parents=True, exist_ok=True)

    # download images and unzip
    key = (DLCModel & key).fetch1("evaluated_frames", download_path=folder)
    zip_path = Path(key)
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(folder)

    return folder


def delete_masks(key, masks, tables):
    """
    Computes masks to be deleted and permanently deletes them
    """

    to_delete = masks[masks["maskfunction_deleted"]]
    to_delete["experimenter"] = key["experimenter"]
    to_delete["behaviourrig_id"] = key["behaviourrig_id"]
    del to_delete["maskfunction_deleted"]
    del to_delete["mask_function"]
    del to_delete["maskfunction_description"]

    (tables["MaskFunction"] & to_delete).delete_quick()


def export_nwb(session, download_path):
    """
    Function exports the NWB file for a session
    """
    assert set(session.keys()) == {
        "experimenter",
        "experiment_id",
        "session_id",
    }, "Function takes session primary key only"

    # function runs in different process so needs to establish its own connection
    if check_streamlit():
        conn_insert = connect.connect()
        tables_insert = import_schemas.schema(conn_insert)
    elif hasattr(sys, "ps1") or sys.flags.interactive:
        import antelope.scripts.hold_conn as hold

        conn_insert = hold.conn
        tables_insert = hold.tables
    else:
        import antelope.load_connection as db

        conn_insert = getattr(db, "conn")
        tables_insert = {}
        for table in dir(db):
            if not table.startswith("_"):
                tables_insert[table] = getattr(db, table)

    # pull session info
    session_info = (
        tables_insert["Session"]
        * tables_insert["Experiment"]
        * tables_insert["Experimenter"]
        & session
    ).fetch1()

    # make nwb root
    nwbfile = NWBFile(
        session_description=session_info["session_notes"],
        identifier=str(uuid4()),
        session_id=f'{session_info["experimenter"]}_{session_info["experiment_id"]}_{session_info["session_id"]}',
        session_start_time=session_info["session_timestamp"],
        experimenter=session_info["full_name"],
        lab=session_info["group"],
        institution=session_info["institution"],
        experiment_description=session_info["experiment_notes"],
    )

    # pull all animals - can be from behaviour or ephys
    query = (
        tables_insert["Animal"]
        & (tables_insert["Self"] * tables_insert["Session"] & session)
    ).proj() + (
        tables_insert["Animal"]
        & (tables_insert["Recording"] * tables_insert["Session"] & session)
    ).proj()
    keys = query.fetch(as_dict=True)
    animals = (tables_insert["Animal"] & keys).fetch(as_dict=True)

    for i, animal in enumerate(animals):
        subject = Subject(
            subject_id=str(animal["animal_id"]),
            species="Mus Musculus",
            age=animal["age"],
            description=animal["animal_notes"],
            sex=animal["sex"],
            strain=animal["genotype"],
        )
        if len(animal) == 1:
            nwbfile.subject = subject
        else:
            setattr(nwbfile, f"subject_{i}", subject)

    # start context manager and start writing to file
    # context manager simply deletes file if there's an error to avoid partially written nwb
    # we then want to write to nwb in chunks as datasets can be large
    with SafeNWBContext(download_path) as f:
        with NWBHDF5IO(f, "w") as io:
            io.write(nwbfile)

        # ephys data
        recs = (
            (tables_insert["Recording"] & session & animals).proj().fetch(as_dict=True)
        )
        for rec in recs:
            pass

            # add probe
            # issue with contact id
            """
            probe = (tables_insert['Animal'] * tables_insert['ProbeInsertion'] * tables_insert['ProbeGeometry'] & rec).fetch1('probe')
            with tempfile.NamedTemporaryFile('w', delete=False, suffix='.json') as tmpfile:
                json.dump(probe, tmpfile)
                path = tmpfile.name
            probefile = pi.read_probeinterface(path)
            ndx_probe = ndx_probeinterface.from_probeinterface(probefile.probes[0])
            nwbfile.add_device(ndx_probe)
            """

        # behaviour data
        with NWBHDF5IO(f, "r+") as io:
            nwbfile = io.read()
            behavior_module = nwbfile.create_processing_module(
                name="behavior", description="behavioral data"
            )
            io.write(nwbfile)

        # analog events
        analog_events = (
            (tables_insert["AnalogEvents"] & session).proj().fetch(as_dict=True)
        )
        for analog_event in analog_events:
            data = (
                tables_insert["AnalogEvents"] * tables_insert["Feature"] & analog_event
            ).fetch1()
            timeseries = TimeSeries(
                name=data["analogevents_name"],
                data=data["data"],
                timestamps=data["timestamps"],
                description=data["feature_description"],
                unit=data["unit"],
            )
            with NWBHDF5IO(f, "r+") as io:
                nwbfile = io.read()
                if data["source_type"] == "processing":
                    behaviour_timeseries = BehavioralTimeSeries(
                        time_series=timeseries, name=data["analogevents_name"]
                    )
                    nwbfile.processing["behavior"].add(behaviour_timeseries)
                elif data["source_type"] == "stimulus":
                    nwbfile.add_stimulus(timeseries)
                elif data["source_type"] == "acquisition":
                    nwbfile.add_acquisition(timeseries)
                io.write(nwbfile)

        # digital events
        digital_events = (
            (tables_insert["DigitalEvents"] & session).proj().fetch(as_dict=True)
        )
        for digital_event in digital_events:
            data = (
                tables_insert["DigitalEvents"] * tables_insert["Feature"]
                & digital_event
            ).fetch1()
            timeseries = TimeSeries(
                name=data["digitalevents_name"],
                data=data["data"],
                timestamps=data["timestamps"],
                description=data["feature_description"],
                unit=data["unit"],
            )
            with NWBHDF5IO(f, "r+") as io:
                nwbfile = io.read()
                if data["source_type"] == "processing":
                    behaviour_timeseries = BehavioralTimeSeries(
                        time_series=timeseries, name=data["digitalevents_name"]
                    )
                    nwbfile.processing["behavior"].add(behaviour_timeseries)
                elif data["source_type"] == "stimulus":
                    nwbfile.add_stimulus(timeseries)
                elif data["source_type"] == "acquisition":
                    nwbfile.add_acquisition(timeseries)
                io.write(nwbfile)

        # interval events
        interval_events = (
            (tables_insert["IntervalEvents"] & session).proj().fetch(as_dict=True)
        )
        for interval_event in interval_events:
            data = (
                tables_insert["IntervalEvents"] * tables_insert["Feature"]
                & interval_event
            ).fetch1()
            timeseries = IntervalSeries(
                name=data["intervalevents_name"],
                data=data["data"],
                timestamps=data["timestamps"],
                description=data["feature_description"],
            )
            with NWBHDF5IO(f, "r+") as io:
                nwbfile = io.read()
                if data["source_type"] == "processing":
                    behaviour_timeseries = BehavioralTimeSeries(
                        time_series=timeseries, name=data["intervalevents_name"]
                    )
                    nwbfile.processing["behavior"].add(behaviour_timeseries)
                elif data["source_type"] == "stimulus":
                    nwbfile.add_stimulus(timeseries)
                elif data["source_type"] == "acquisition":
                    nwbfile.add_acquisition(timeseries)
                io.write(nwbfile)

        # interval events
        kinematics = (tables_insert["Kinematics"] & session).proj().fetch(as_dict=True)
        with NWBHDF5IO(f, "r+") as io:
            nwbfile = io.read()
            position = Position()
            nwbfile.processing["behavior"].add(position)
            io.write(nwbfile)
        for kinematic in kinematics:
            rig_json = (tables_insert["BehaviourRig"] & kinematic).fetch1("rig_json")
            data = (
                tables_insert["Kinematics"] * tables_insert["Feature"] & kinematic
            ).fetch1()
            timeseries = SpatialSeries(
                name=data["kinematics_name"],
                data=data["data"],
                timestamps=data["timestamps"],
                description=data["feature_description"],
                reference_frame=rig_json["reference_point"],
            )
            with NWBHDF5IO(f, "r+") as io:
                nwbfile = io.read()
                nwbfile.processing["behavior"]["Position"].add_spatial_series(
                    timeseries
                )
                io.write(nwbfile)

        # masks
        mask_epochs = BehavioralEpochs(name="masks")
        masks = (tables_insert["Mask"] & session).proj().fetch(as_dict=True)
        for mask in masks:
            data = (
                tables_insert["Mask"] * tables_insert["MaskFunction"] & mask
            ).fetch1()
            timeseries = IntervalSeries(
                name=data["mask_name"],
                data=data["data"],
                timestamps=data["timestamps"],
                description=data["maskfunction_description"],
            )
            mask_epochs.add_interval_series(timeseries)
        with NWBHDF5IO(f, "r+") as io:
            nwbfile = io.read()
            nwbfile.processing["behavior"].add(mask_epochs)
            io.write(nwbfile)

        # videos
        videos = (tables_insert["Video"] & session).proj().fetch(as_dict=True)
        for video in videos:
            rig_json = (tables_insert["BehaviourRig"] & video).fetch1("rig_json")
            video_name = rig_json["videos"][video["video_id"] - 1]["name"]
            video_description = rig_json["videos"][video["video_id"] - 1]["description"]
            data = (tables_insert["Video"] & video).fetch1(
                download_path=Path(download_path).parent
            )
            behaviour_video = ImageSeries(
                name=video_name,
                external_file=[data["video"]],
                starting_frame=[0],
                format="external",
                timestamps=data["timestamps"],
                unit="s",
                description=video_description,
            )
            with NWBHDF5IO(f, "r+") as io:
                nwbfile = io.read()
                nwbfile.add_acquisition(behaviour_video)
                io.write(nwbfile)


class SafeNWBContext:
    """
    Context manager deletes NWB file if an error occurs
    Designed so that the file should be opened within this context, several times if necessary for memory reasons.
    """

    def __init__(self, file_path):
        self.file_path = file_path

    def __enter__(self):
        return self.file_path

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            print(f"Exception occurred: {exc_value}")
            traceback.print_tb(exc_traceback)

            if os.path.exists(self.file_path):
                os.remove(self.file_path)
                print(f"File {self.file_path} deleted due to error.")



import hashlib
import json
import numpy as np
import inspect
from pathlib import Path


def hash_list(data_list):
    # Serialize the list to a JSON string
    json_string = json.dumps(data_list)

    # Encode the JSON string to bytes
    json_bytes = json_string.encode("utf-8")

    # Create a SHA-256 hash object
    sha256 = hashlib.sha256()

    # Update the hash object with the byte representation
    sha256.update(json_bytes)

    # Get the hexadecimal representation of the hash
    hash_hex = sha256.hexdigest()

    return hash_hex


def hash_query(query):
    """
    Function produces a reproducible hash of a query.
    It uses the MD5 algorithm on the database side and is therefore quite fast.
    """

    hashtable = query.proj(
        row_hash=f"MD5(CONCAT_WS('|',{', '.join(f'HEX(`{k}`)' for k in query.heading)}))"
    ).fetch("row_hash")

    return hashtable


def hash_tables(tables, key):
    """
    Function hashes the data across several tables belonging to a given key.
    This can be used to check if data is consistent fo a given key across the database.

    Arguments:
    tables: list of tables to hash
    key: dict giving a datajoint restriction
    """

    table_hashes = []
    for table in tables:
        query = table & key
        row_hashes = hash_query(query)
        row_hashes = list(np.sort(row_hashes))
        table_hashes.append(hash_list(row_hashes))

    hash = hash_list(table_hashes)
    return hash


def code_hash(function):
    """
    Function hashes the code of a function.
    """

    code = inspect.getsource(function.run)
    code = code.encode("utf-8")
    sha256 = hashlib.sha256()
    sha256.update(code)
    hash = sha256.hexdigest()
    return hash


def list_to_datajoint(tables, tablenames):
    """
    Function to convert a list of tablenames to a list of datajoint tables.
    """

    table_list = []
    for table in tablenames:
        table_list.append(tables[table])
    return table_list


def dfs_functions(self, functions):
    """
    Does a depth first search to find the graph of all functions that are called recursively.
    Assumes functions are called in a tree. Could be modified if loops are allowed later.
    """

    # base case - return tables
    if not hasattr(self, "calls"):
        tables = []
        if isinstance(self.query, str):
            tables.append(self.query)
        elif isinstance(self.query, list):
            tables = self.query
        if hasattr(self, "data"):
            if isinstance(self.data, str):
                tables.append(self.data)
            elif isinstance(self.data, list):
                tables += self.data
        return tables

    # otherwise recurse
    else:

        tables = []
        # loop through all recursed functions
        for function in self.calls:

            # actually instantiate function from string
            if "." in function:
                folder, function = function.split(".")
                for f in functions[folder]:
                    if f.name == function:
                        function_inst = f
            else:
                folder = Path(inspect.getsourcefile(self.__class__.__bases__[0])).stem
                for f in functions[folder]:
                    if f.name == function:
                        function_inst = f

            # recurse, accumulating children tables
            tables.extend(dfs_functions(function_inst, functions))

        # append self's tables
        if isinstance(self.query, str):
            tables.append(self.query)
        elif isinstance(self.query, list):
            tables.extend(self.query)
        if hasattr(self, "data"):
            if isinstance(self.data, str):
                tables.append(self.data)
            elif isinstance(self.data, list):
                tables.extend(self.data)
        tables = list(set(tables))

        return tables


def dfs_functions_not_tables(main, functions, stored=[]):
    """
    Just returns all functions that are called recursively.
    """

    # base case
    if not hasattr(main, "calls"):
        return []

    # otherwise recurse
    else:

        main.folder = Path(inspect.getsourcefile(main.__class__.__bases__[0])).stem
        stored.append(main)

        # loop through all recursed functions
        for function in main.calls:
            if function not in stored:
                # actually instantiate function from string
                if "." in function:
                    folder, function = function.split(".")
                    for f in functions[folder]:
                        if f.name == function:
                            function_inst = f
                            function_inst.folder = folder
                else:
                    folder = Path(
                        inspect.getsourcefile(main.__class__.__bases__[0])
                    ).stem
                    for f in functions[folder]:
                        if f.name == function:
                            function_inst = f
                            function_inst.folder = folder
                stored.append(function_inst)

                # recurse, accumulating children tables
                stored.extend(
                    dfs_functions_not_tables(function_inst, functions, stored)
                )

        return stored



import streamlit as st
import streamlit_antd_components as sac
import os
from pathlib import Path


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:
        # query user's full name from database
        @st.cache_data
        def name(username):
            query = tables["Experimenter"] & {"experimenter": username}
            name = query.fetch1("full_name")
            return name

        name = name(username)

        # display welcome
        st.title("Antelope")
        st.write(f"Welcome back {name.split()[0]}!")
        st.write(
            """If you haven't done so already, please take a look at our documentation at:"""
        )
        st.write("https://antelope.readthedocs.io")

        st.divider()

        # display database structure
        st.header("Database structure")

        schema = ["session", "ephys", "behaviour"][
            sac.segmented(
                items=[
                    sac.SegmentedItem(label="Session", icon="table"),
                    sac.SegmentedItem(label="Electrophysiology", icon="lightning"),
                    sac.SegmentedItem(label="Behaviour", icon="joystick"),
                ],
                use_container_width=True,
                return_index=True,
            )
        ]

        resources = Path(os.path.abspath(__file__)).parent.parent / "resources"
        relative_resources = Path(
            os.path.relpath(resources, Path.cwd())
        )  # needed because of windows image bug
        image = relative_resources / f"{schema}.png"

        st.image(str(image), use_column_width=True)

        st.divider()

        # display general usage info
        st.header("Usage guide")
        st.write(
            "First of all, make sure you are familiar with the database structures shown above."
        )
        st.write(
            """Crucially, the tables in green are **Manual** tables, which require user input.
                The tables in red are **Computed** tables, which are populated automatically by submitting computations to the HPC.
                The tables in grey are **Lookup** tables, which are populated by the database administrators.
                The tables in blue are **Imported** tables, which are populated by importing data external data.
                """
        )
        st.write("Use the sidebar on the left to navigate between pages.")
        st.markdown(
            """
    * The **search** page allows you to search a database table, and download or delete the data.
    * The **schemas** tab guides you through uploading data to the various database schemas, perform computations, and visualise the results.
    * The **visualisation** tab allows you to interactively visualise the data in the database.
    * The **analysis** tab allows you to perform preliminary analysis on the data in the database.
    * The **admin** page (only available to database administrators) allows you to search deleted entries, and either restore them, or permanently delete them.
    """
        )
        st.write(
            "Please familiarise yourself with the documentation before using antelope."
        )
